"""Projection Assumptions


This Space associates projection assumptions to model points,
by looking up paramters in a table stored as an `ExcelRange`_ object
associated to a Reference named :attr:`Assumption`.

.. _PandasData:
   https://docs.modelx.io/en/latest/reference/dataclient.html#pandasdata

.. _ExcelRange:
   https://docs.modelx.io/en/latest/reference/dataclient.html#excelrange


.. rubric:: References

Attributes:
    Assumption: `ExcelRange`_ object for associating
        assumption identifiers to model point keys
    AssumptionTables: `ExcelRange`_ object holding assumption tables
        such as surrender rates and mortality factors
    PolicyData: `PandasData`_ object holding model point data as a pandas
        DataFrame read from the model point file
    TableLastAge: Reference to :func:`fastlife.model.Input.TableLastAge`
    prod: Reference to :func:`fastlife.Projection.Policy.Product`
    polt: Reference to :func:`fastlife.Projection.Policy.PolicyType`
    gen: Reference to :func:`fastlife.Projection.Policy.Gen`
    sex: Reference to :func:`fastlife.Projection.Policy.Sex`

"""

from modelx.serialize.jsonvalues import *

_formula = None

_bases = []

_allow_none = True

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def CnsmpTax():
    """Consumption tax rate"""
    return AsmpLookup("CnsmpTax")


def CommInitPrem():
    """Initial commission per premium"""

    def get_table_id(pol):
        key1, key2, key3 = pol['Product'], pol['PolicyType'], pol['Gen']
        result = AsmpLookup.match("CommInitPrem", key1, key2, key2).value

        if result is not None:
            return result
        else:
            raise ValueError('CommInitPrem not found')

    return PolicyData().apply(get_table_id, axis=1)


def CommRenPrem():
    """Renewal commission per premium"""

    def get_table_id(pol):
        key1, key2, key3 = pol['Product'], pol['PolicyType'], pol['Gen']
        result = AsmpLookup.match("CommRenPrem", key1, key2, key2).value

        if result is not None:
            return result
        else:
            raise ValueError('CommRenPrem not found')

    return PolicyData().apply(get_table_id, axis=1)


def CommRenTerm():
    """Renewal commission term"""

    def get_table_id(pol):
        key1, key2, key3 = pol['Product'], pol['PolicyType'], pol['Gen']
        result = AsmpLookup.match("CommRenTerm", key1, key2, key2).value

        if result is not None:
            return result
        else:
            raise ValueError('CommRenTerm not found')

    return PolicyData().apply(get_table_id, axis=1)


def ExpsAcqAnnPrem():
    """Acquisition expense per annualized premium"""

    def get_table_id(pol):
        key1, key2, key3 = pol['Product'], pol['PolicyType'], pol['Gen']
        return AsmpLookup.match("ExpsAcqAnnPrem", key1, key2, key2).value

    return PolicyData().apply(get_table_id, axis=1)


def ExpsAcqPol():
    """Acquisition expense per policy"""

    def get_table_id(pol):
        key1, key2, key3 = pol['Product'], pol['PolicyType'], pol['Gen']
        return AsmpLookup.match("ExpsAcqPol", key1, key2, key2).value

    return PolicyData().apply(get_table_id, axis=1)


def ExpsAcqSA():
    """Acquisition expense per sum assured"""

    def get_table_id(pol):
        key1, key2, key3 = pol['Product'], pol['PolicyType'], pol['Gen']
        return AsmpLookup.match("ExpsAcqSA", key1, key2, key2).value

    return PolicyData().apply(get_table_id, axis=1)


def ExpsMaintAnnPrem():
    """Maintenance expense per annualized premium"""

    def get_table_id(pol):
        key1, key2, key3 = pol['Product'], pol['PolicyType'], pol['Gen']
        return AsmpLookup.match("ExpsMaintPrem", key1, key2, key2).value

    return PolicyData().apply(get_table_id, axis=1)


def ExpsMaintPol():
    """Maintenance expense per policy"""

    def get_table_id(pol):
        key1, key2, key3 = pol['Product'], pol['PolicyType'], pol['Gen']
        return AsmpLookup.match("ExpsMaintPol", key1, key2, key2).value

    return PolicyData().apply(get_table_id, axis=1)


def ExpsMaintSA():
    """Maintenance expense per sum assured"""


    def get_table_id(pol):
        key1, key2, key3 = pol['Product'], pol['PolicyType'], pol['Gen']
        return AsmpLookup.match("ExpsMaintSA", key1, key2, key2).value

    return PolicyData().apply(get_table_id, axis=1)


def InflRate():
    """Inflation rate"""
    return AsmpLookup("InflRate")


def LastAge():
    """Age at which mortality becomes 1"""

    keys = pd.concat([MortTableID(), sex()], axis=1)

    result = keys.join(TableLastAge(), on=['MortTableID', 'Sex'])['TableLastAge']

    result.name = 'LastAge'

    return result


def MortFactor(y):
    """Mortality factor"""

    fac = MortFactorID().apply(lambda facid: AssumptionTables.get((facid, y), np.NaN))

    if y == 0 or not fac.isnull().any():
        return fac
    else:
        return fac.mask(fac.isnull(), MortFactor(y-1))


def SurrRate(y):
    """Surrender Rate"""

    fac = SurrRateID().apply(lambda surrid: AssumptionTables.get((surrid, y), np.NaN))

    if y == 0 or not fac.isnull().any():
        return fac
    else:
        return fac.mask(fac.isnull(), SurrRate(y-1))


def MortTableID():
    """Mortality Table"""

    def get_table_id(pol):
        key1, key2, key3 = pol['Product'], pol['PolicyType'], pol['Gen']
        return AsmpLookup.match("BaseMort", key1, key2, key2).value

    result = PolicyData().apply(get_table_id, axis=1)
    result.name = 'MortTableID'

    return result


def MortFactorID():
    """Mortality factor"""

    def get_factor(pol):

        key1, key2, key3 = pol['Product'], pol['PolicyType'], pol['Gen']
        table = AsmpLookup.match("MortFactor", key1, key2, key2).value

        if table is None:
            raise ValueError('MortFactor not found')

        return table

    return PolicyData().apply(lambda pol: get_factor(pol), axis=1)


def SurrRateID():

    def get_factor(pol):

        key1, key2, key3 = pol['Product'], pol['PolicyType'], pol['Gen']
        table = AsmpLookup.match("Surrender", key1, key2, key2).value

        if table is None:
            raise ValueError('Surrender not found')

        return table

    return PolicyData().apply(lambda pol: get_factor(pol), axis=1)


def AsmpLookup(asmp, prod=None, polt=None, gen=None):
    """Look up assumptions"""
    return Assumption.get((asmp, prod, polt, gen), None)


# ---------------------------------------------------------------------------
# References

Assumption = ("Pickle", 3020548391688)

AssumptionTables = ("Pickle", 3020548383112)

PolicyData = ("Pickle", 3020548337160)

TableLastAge = ("Interface", ("...", "Input", "TableLastAge"), "auto")

gen = ("Interface", ("..", "Policy", "Gen"), "auto")

polt = ("Interface", ("..", "Policy", "PolicyType"), "auto")

prod = ("Interface", ("..", "Policy", "Product"), "auto")

sex = ("Interface", ("..", "Policy", "Sex"), "auto")

MortalityTables = ("Pickle", 3020541952136)


from modelx.serialize.jsonvalues import *

_formula = lambda PolicyID, ScenID=1: None

_bases = [
    ".IFRS",
    ".BaseProj"
]

_allow_none = None

_spaces = [
    "InnerProj",
    "Policy",
    "Assumptions"
]

# ---------------------------------------------------------------------------
# Cells

def IntAccumCF(t):
    """Intrest on accumulated cashflows"""
    return (AccumCF(t)
            + PremIncome(t)
            - ExpsTotal(t)) * DiscRate(t, 0)


def DiscRate(t, dur):
    """Discount rates for the outer projection"""
    return scen.DiscRate(dur) + DiscRateAdj(t)


def DiscRateAdj(t):
    """Adjustment to the outer discount rates"""
    if t == 0:
        return 0
    else:
        return DiscRateAdj(t-1)


# ---------------------------------------------------------------------------
# References

pol = ("Interface", (".", "Policy"), "auto")


import modelx as mx
import lifelib.projects.simplelife.model
import lifelib.projects.fastlife.model


import pathlib

simplepath = pathlib.Path(lifelib.projects.simplelife.model.__file__).parent
fastpath = pathlib.Path(lifelib.projects.fastlife.model.__file__).parent


def test_fastlife():

    simple = mx.read_model(simplepath)
    fast = mx.read_model(fastpath)

    simplesum = sum([simple.Projection[i].PV_NetCashflow(0) for i in range(1, 301)])
    fastsum = sum(fast.Projection.PV_NetCashflow(0))

    assert simplesum == fastsum



import sys
import os.path
import pickle
import pathlib

import pytest

from lifelib.projects.nestedlife.scripts import nestedlife
from lifelib.tests.data.generate_testdata_nestedlife_v0 import (
    round_signif,
    get_nested,
    set_model,
    update_model)

if '' not in sys.path:
    sys.path.insert(0, '')


datadir = pathlib.Path(__file__).parents[1].joinpath('data')

testdata1 = str(datadir.joinpath('data_nestedlife1'))
testdata2 = str(datadir.joinpath('data_nestedlife2'))


@pytest.mark.parametrize("testdata, func",[
    [testdata1, set_model],
    [testdata2, update_model]
])
def test_nestedlife(testdata, func):
    model = nestedlife.build(load_saved=False)

    data = get_nested(func(model), 'PolsSurr')

    with open(testdata, 'rb') as file:
        data_saved = pickle.load(file)

    assert data == data_saved





"""The main Space in the :mod:`~basiclife.BasicTerm_SE` model.

:mod:`~basiclife.BasicTerm_SE.Projection` is the only Space defined
in the :mod:`~basiclife.BasicTerm_SE` model, and it contains
all the logic and data used in the model.

.. rubric:: Parameters and References

(In all the sample code below,
the global variable ``Projection`` refers to the
:mod:`~basiclife.BasicTerm_SE.Projection` Space.)

Attributes:

    point_id: The ID of the selected model point.
        ``point_id`` is defined as a Reference, and its value
        is used for determining the selected model point.
        By default, ``1`` is assigned. To select another model point,
        assign its model point ID to it::

            >>> Projection.point_id = 2

        ``point_id`` is also defined as the parameter of the
        :mod:`~basiclife.BasicTerm_SE.Projection` Space,
        which makes it possible to create dynamic child space
        for multiple model points::

            >>> Projection.parameters
            ('point_id',)

            >>> Projection[1]
            <ItemSpace BasicTerm_SE.Projection[1]>

            >>> Projection[2]
            <ItemSpace BasicTerm_SE.Projection[2]>

        .. seealso::

           * :attr:`model_point_table`
           * :func:`model_point`

    model_point_table: All model points as a DataFrame.
        The sample model point data was generated by
        *generate_model_points_with_duration.ipynb* included in the library.
        The DataFrame has an index named ``point_id``,
        and :func:`model_point` returns a record as a Series
        whose index value matches :attr:`point_id`.
        The DataFrame has the following columns:

            * ``age_at_entry``
            * ``sex``
            * ``policy_term``
            * ``policy_count``
            * ``sum_assured``
            * ``duration_mth``

        Cells defined in :mod:`~basiclife.BasicTerm_SE.Projection`
        with the same names as these columns return
        the corresponding column's values for the selected model point.

        .. code-block::

            >>> Projection.model_poit_table
                       age_at_entry sex  ...  sum_assured  duration_mth
            policy_id                    ...
            1                    47   M  ...       622000             1
            2                    29   M  ...       752000           210
            3                    51   F  ...       799000            15
            4                    32   F  ...       422000           125
            5                    28   M  ...       605000            55
                            ...  ..  ...          ...           ...
            9996                 47   M  ...       827000           157
            9997                 30   M  ...       826000           168
            9998                 45   F  ...       783000           146
            9999                 39   M  ...       302000            11
            10000                22   F  ...       576000           166

            [10000 rows x 6 columns]

        The DataFrame is saved in the Excel file *model_point_table.xlsx*
        placed in the model folder.
        :attr:`model_point_table` is created by
        Projection's `new_pandas`_ method,
        so that the DataFrame is saved in the separate file.
        The DataFrame has the injected attribute
        of ``_mx_dataclident``::

            >>> Projection.model_point_table._mx_dataclient
            <PandasData path='model_point_table.xlsx' filetype='excel'>

        .. seealso::

           * :attr:`point_id`
           * :func:`model_point`
           * :func:`age_at_entry`
           * :func:`sex`
           * :func:`policy_term`
           * :func:`pols_if_init`
           * :func:`sum_assured`
           * :func:`duration_mth`

    premium_table: Premium rate table by entry age and duration as a Series.
        The table is created using :mod:`~basiclife.BasicTerm_M`
        as demonstrated in *create_premium_table.ipynb*.
        The table is stored in *premium_table.xlsx* in the model folder.

        .. code-block::

            >>> Projection.premium_table
            age_at_entry  policy_term
            20            10             0.000046
                          15             0.000052
                          20             0.000057
            21            10             0.000048
                          15             0.000054
                                           ...
            58            15             0.000433
                          20             0.000557
            59            10             0.000362
                          15             0.000471
                          20             0.000609
            Name: premium_rate, Length: 120, dtype: float64

    disc_rate_ann: Annual discount rates by duration as a pandas Series.

        .. code-block::

            >>> Projection.disc_rate_ann
            year
            0      0.00000
            1      0.00555
            2      0.00684
            3      0.00788
            4      0.00866

            146    0.03025
            147    0.03033
            148    0.03041
            149    0.03049
            150    0.03056
            Name: disc_rate_ann, Length: 151, dtype: float64

        The Series is saved in the Excel file *disc_rate_ann.xlsx*
        placed in the model folder.
        :attr:`disc_rate_ann` is created by
        Projection's `new_pandas`_ method,
        so that the Series is saved in the separate file.
        The Series has the injected attribute
        of ``_mx_dataclident``::

            >>> Projection.disc_rate_ann._mx_dataclient
            <PandasData path='disc_rate_ann.xlsx' filetype='excel'>

        .. seealso::

           * :func:`disc_rate_mth`
           * :func:`disc_factors`

    mort_table: Mortality table by age and duration as a DataFrame.
        See *basic_term_sample.xlsx* included in this library
        for how the sample mortality rates are created.

        .. code-block::

            >>> Projection.mort_table
                        0         1         2         3         4         5
            Age
            18   0.000231  0.000254  0.000280  0.000308  0.000338  0.000372
            19   0.000235  0.000259  0.000285  0.000313  0.000345  0.000379
            20   0.000240  0.000264  0.000290  0.000319  0.000351  0.000386
            21   0.000245  0.000269  0.000296  0.000326  0.000359  0.000394
            22   0.000250  0.000275  0.000303  0.000333  0.000367  0.000403
            ..        ...       ...       ...       ...       ...       ...
            116  1.000000  1.000000  1.000000  1.000000  1.000000  1.000000
            117  1.000000  1.000000  1.000000  1.000000  1.000000  1.000000
            118  1.000000  1.000000  1.000000  1.000000  1.000000  1.000000
            119  1.000000  1.000000  1.000000  1.000000  1.000000  1.000000
            120  1.000000  1.000000  1.000000  1.000000  1.000000  1.000000

            [103 rows x 6 columns]

        The DataFrame is saved in the Excel file *mort_table.xlsx*
        placed in the model folder.
        :attr:`mort_table` is created by
        Projection's `new_pandas`_ method,
        so that the DataFrame is saved in the separate file.
        The DataFrame has the injected attribute
        of ``_mx_dataclident``::

            >>> Projection.mort_table._mx_dataclient
            <PandasData path='mort_table.xlsx' filetype='excel'>

        .. seealso::

           * :func:`mort_rate`
           * :func:`mort_rate_mth`

    np: The `numpy`_ module.
    pd: The `pandas`_ module.

.. _numpy:
   https://numpy.org/

.. _pandas:
   https://pandas.pydata.org/

.. _new_pandas:
   https://docs.modelx.io/en/latest/reference/space/generated/modelx.core.space.UserSpace.new_pandas.html

"""

from modelx.serialize.jsonvalues import *

_formula = lambda point_id: None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def age(t):
    """The attained age at time t.

    Defined as::

        age_at_entry() + duration(t)

    .. seealso::

        * :func:`age_at_entry`
        * :func:`duration`

    """
    return age_at_entry() + duration(t)


def age_at_entry():
    """The age at entry of the selected model point

    The element labeled ``age_at_entry`` of the Series returned by
    :func:`model_point`.
    """
    return model_point()["age_at_entry"]


def check_pv_net_cf():
    """Check present value summation

    Check if the present value of :func:`net_cf` matches the
    sum of the present values each cashflows.
    Returns the check result as :obj:`True` or :obj:`False`.

     .. seealso::

        * :func:`net_cf`
        * :func:`pv_net_cf`

    """

    import math
    res = sum(list(net_cf(t) for t in range(proj_len())) * disc_factors()[:proj_len()])

    return math.isclose(res, pv_net_cf())


def claim_pp(t):
    """Claim per policy

    The claim amount per plicy. Defaults to :func:`sum_assured`.
    """
    return sum_assured()


def claims(t):
    """Claims

    Claims during the period from ``t`` to ``t+1`` defined as::

        claim_pp(t) * pols_death(t)

    .. seealso::

        * :func:`claim_pp`
        * :func:`pols_death`

    """
    return claim_pp(t) * pols_death(t)


def commissions(t): 
    """Commissions

    By default, 100% premiums for the first year, 0 otherwise.

    .. seealso::

        * :func:`premiums`
        * :func:`duration`

    """
    return premiums(t) if duration(t) == 0 else 0


def disc_factors():
    """Discount factors.

    Vector of the discount factors as a Numpy array. Used for calculating
    the present values of cashflows.

    .. seealso::

        :func:`disc_rate_mth`
    """
    return np.array(list((1 + disc_rate_mth()[t])**(-t) for t in range(proj_len())))


def disc_rate_mth():
    """Monthly discount rate

    Nummpy array of monthly discount rates from time 0 to :func:`proj_len` - 1
    defined as::

        (1 + disc_rate_ann)**(1/12) - 1

    .. seealso::

        :func:`disc_rate_ann`

    """
    return np.array(list((1 + disc_rate_ann[t//12])**(1/12) - 1 for t in range(proj_len())))


def duration(t):
    """Duration of the selected model point at ``t`` in years

    .. seealso:: :func:`duration_mth`

    """
    return duration_mth(t) // 12


def duration_mth(t):
    """Duration of the selected model point at ``t`` in months

    Indicates how many months the policies have been in-force at ``t``.
    The initial value at time 0 is read from the ``duration_mth`` column in
    :attr:`model_point_table` through :func:`model_point`.
    Increments by 1 as ``t`` increments.
    Negative values of :func:`duration_mth` indicate future new business
    policies. For example, If the :func:`duration_mth` is
    -15 at time 0, the model point is issued at ``t=15``.

    .. seealso:: :func:`model_point`

    """
    if t == 0:
        return model_point()['duration_mth']
    else:
        return duration_mth(t-1) + 1


def expense_acq():
    """Acquisition expense per policy

    ``300`` by default.
    """
    return 300


def expense_maint():
    """Annual maintenance expense per policy

    ``60`` by default.
    """
    return 60


def expenses(t):
    """Expenses

    Expenses during the period from ``t`` to ``t+1``
    defined as the sum of acquisition expenses and maintenance expenses.
    The acquisition expenses are modeled as :func:`expense_acq`
    times :func:`pols_new_biz`.
    The maintenance expenses are modeled as :func:`expense_maint`
    times :func:`inflation_factor` times :func:`pols_if_at` before
    decrement.

    .. seealso::

        * :func:`expense_acq`
        * :func:`expense_maint`
        * :func:`inflation_factor`
        * :func:`pols_new_biz`
        * :func:`pols_if_at`

    """

    return expense_acq() * pols_new_biz(t) \
        + pols_if_at(t, "BEF_DECR") * expense_maint()/12 * inflation_factor(t)


def inflation_factor(t):
    """The inflation factor at time t

    .. seealso::

        * :func:`inflation_rate`

    """
    return (1 + inflation_rate)**(t/12)


def inflation_rate():
    """Inflation rate"""
    return 0.01


def lapse_rate(t):
    """Lapse rate

    By default, the lapse rate assumption is defined by duration as::

        max(0.1 - 0.02 * duration(t), 0.02)

    .. seealso::

        :func:`duration`

    """
    return max(0.1 - 0.02 * duration(t), 0.02)


def loading_prem():
    """Loading per premium

    .. note::
       This cells is not used by default.

    ``0.5`` by default.

    .. seealso::

        * :func:`premium_pp`

    """
    return 0.50


def model_point():
    """The selected model point as a Series

    :func:`model_point` looks up :attr:`model_point_table`, and
    returns as a Series the row whose index is the value of
    :attr:`point_id`.

    Example:
        In the code below ``Projection`` refers to
        the :mod:`~basiclife.BasicTerm_SE.Projection` space::

            >>> Projection.point_id
            1

            >>> Projection.model_point()
            age_at_entry        47
            sex                  M
            policy_term         10
            policy_count        86
            sum_assured     622000
            duration_mth         1
            Name: 1, dtype: object

            >>> Projection.point_id = 2

            >>> Projection.model_point()
            age_at_entry        29
            sex                  M
            policy_term         20
            policy_count        56
            sum_assured     752000
            duration_mth       210
            Name: 2, dtype: object

    """
    return model_point_table.loc[point_id]


def mort_rate(t):
    """Mortality rate to be applied at time t

    .. seealso::

       * :attr:`mort_table`
       * :func:`mort_rate_mth`

    """
    return mort_table[str(max(min(5, duration(t)),0))][age(t)]


def mort_rate_mth(t):
    """Monthly mortality rate to be applied at time t

    .. seealso::

       * :attr:`mort_table`
       * :func:`mort_rate`

    """
    return 1-(1- mort_rate(t))**(1/12)


def net_cf(t):
    """Net cashflow

    Net cashflow for the period from ``t`` to ``t+1`` defined as::

        premiums(t) - claims(t) - expenses(t) - commissions(t)

    .. seealso::

        * :func:`premiums`
        * :func:`claims`
        * :func:`expenses`
        * :func:`commissions`

    """
    return premiums(t) - claims(t) - expenses(t) - commissions(t)


def net_premium_pp():
    """Net premium per policy

    .. note::
       This cells is not used by default.

    The net premium per policy is defined so that
    the present value of net premiums equates to the present value of
    claims::

        pv_claims() / pv_pols_if()

    .. seealso::

        * :func:`pv_claims`
        * :func:`pv_pols_if`

    """
    return pv_claims() / pv_pols_if()


def policy_term():
    """The policy term of the selected model point.

    The element labeled ``policy_term`` of the Series returned by
    :func:`model_point`.
    """

    return model_point()["policy_term"]


def pols_death(t):
    """Number of death occurring at time t"""
    return pols_if_at(t, "BEF_DECR") * mort_rate_mth(t)


def pols_if(t):
    """Number of policies in-force

    :func:`pols_if(t)<pols_if>` is an alias
    for :func:`pols_if_at(t, "BEF_MAT")<pols_if_at>`.

    .. seealso::
        * :func:`pols_if_at`

    """
    return pols_if_at(t, "BEF_MAT")


def pols_if_at(t, timing):
    """Number of policies in-force

    :func:`pols_if_at(t, timing)<pols_if_at>` calculates
    the number of policies in-force at time ``t``.
    The second parameter ``timing`` takes a string value to
    indicate the timing of in-force,
    which is either
    ``"BEF_MAT"``, ``"BEF_NB"`` or ``"BEF_DECR"``.

    .. rubric:: BEF_MAT

    The number of policies in-force before maturity after lapse and death.
    At time 0, the value is read from :func:`pols_if_init`.
    For time > 0, defined as::

        pols_if_at(t-1, "BEF_DECR") - pols_lapse(t-1) - pols_death(t-1)

    .. rubric:: BEF_NB

    The number of policies in-force before new business after maturity.
    Defined as::

        pols_if_at(t, "BEF_MAT") - pols_maturity(t)

    .. rubric:: BEF_DECR

    The number of policies in-force before lapse and death after new business.
    Defined as::

        pols_if_at(t, "BEF_NB") + pols_new_biz(t)

    .. seealso::
        * :func:`pols_if_init`
        * :func:`pols_lapse`
        * :func:`pols_death`
        * :func:`pols_maturity`
        * :func:`pols_new_biz`
        * :func:`pols_if`

    """
    if timing == "BEF_MAT":

        if t == 0:
            return pols_if_init()
        else:
            return pols_if_at(t-1, "BEF_DECR") - pols_lapse(t-1) - pols_death(t-1)

    elif timing == "BEF_NB":

        return pols_if_at(t, "BEF_MAT") - pols_maturity(t)

    elif timing == "BEF_DECR":

        return pols_if_at(t, "BEF_NB") + pols_new_biz(t)

    else:
        raise ValueError("invalid timing")


def pols_if_init(): 
    """Initial number of policies in-force

    Number of in-force policies at time 0 referenced from
    :func:`pols_if_at(0, "BEF_MAT")<pols_if_at>`.
    """
    if duration_mth(0) > 0:
        return model_point()["policy_count"]
    else:
        return 0


def pols_lapse(t):
    """Number of lapse occurring at time t

    .. seealso::
        * :func:`pols_if_at`
        * :func:`lapse_rate`

    """
    return (pols_if_at(t, "BEF_DECR") - pols_death(t)) * (1-(1 - lapse_rate(t))**(1/12))


def pols_maturity(t):
    """Number of maturing policies

    The policy maturity occurs when
    :func:`duration_mth` equals 12 times :func:`policy_term`.
    The amount is equal to :func:`pols_if_at(t, "BEF_MAT")<pols_if_at>`.

    otherwise ``0``.
    """
    if duration_mth(t) == policy_term() * 12:
        return pols_if_at(t, "BEF_MAT")
    else:
        return 0


def pols_new_biz(t):
    """Number of new business policies

    The number of new business policies.
    The value :func:`duration_mth(0)<duration_mth>`
    for the selected model point is read from the ``policy_count`` column in
    :func:`model_point`. If the value is 0 or negative,
    the model point is new business at t=0 or at t when
    :func:`duration_mth(t)<duration_mth>` is 0, and the
    :func:`pols_new_biz(t)<pols_new_biz>` is read from the ``policy_count``
    in :func:`model_point`.

    .. seealso::
        * :func:`model_point`

    """
    if duration_mth(t) == 0:
        return model_point()['policy_count']
    else:
        return 0


def premium_pp():
    """Monthly premium per policy

    Monthly premium amount per policy defined as::

        round(sum_assured() * premium_table[age_at_entry(), policy_term()], 2)

    .. seealso::

        * :attr:`premium_table`
        * :func:`sum_assured`
        * :func:`age_at_entry`
        * :func:`policy_term`


    """
    return round(sum_assured() * premium_table[age_at_entry(), policy_term()], 2)


def premiums(t):
    """Premium income

    Premium income during the period from ``t`` to ``t+1`` defined as::

        premium_pp() * pols_if_at(t, "BEF_DECR")

    .. seealso::

        * :func:`premium_pp`
        * :func:`pols_if_at`

    """
    return premium_pp() * pols_if_at(t, "BEF_DECR")


def proj_len():
    """Projection length in months

    :func:`proj_len` indicates how many months the projection
    for the selected model point should be carried out. Defined as::

        12 * policy_term() - duration_mth(0) + 1

    .. seealso::

        :func:`policy_term`

    """
    return max(12 * policy_term() - duration_mth(0) + 1, 0)


def pv_claims():
    """Present value of claims

    .. seealso::

        * :func:`claims`

    """
    return sum(list(claims(t) for t in range(proj_len())) * disc_factors()[:proj_len()])


def pv_commissions():
    """Present value of commissions

    .. seealso::

        * :func:`expenses`

    """
    return sum(list(commissions(t) for t in range(proj_len())) * disc_factors()[:proj_len()])


def pv_expenses():
    """Present value of expenses

    .. seealso::

        * :func:`expenses`

    """
    return sum(list(expenses(t) for t in range(proj_len())) * disc_factors()[:proj_len()])


def pv_net_cf():
    """Present value of net cashflows.

    Defined as::

        pv_premiums() - pv_claims() - pv_expenses() - pv_commissions()

    .. seealso::

        * :func:`pv_premiums`
        * :func:`pv_claims`
        * :func:`pv_expenses`
        * :func:`pv_commissions`

    """

    return pv_premiums() - pv_claims() - pv_expenses() - pv_commissions()


def pv_pols_if():
    """Present value of policies in-force

    .. note::
       This cells is not used by default.

    The discounted sum of the number of in-force policies at each month.
    It is used as the annuity factor for calculating :func:`net_premium_pp`.

    """
    return sum(list(pols_if(t) for t in range(proj_len())) * disc_factors()[:proj_len()])


def pv_premiums():
    """Present value of premiums

    .. seealso::

        * :func:`premiums`

    """
    return sum(list(premiums(t) for t in range(proj_len())) * disc_factors()[:proj_len()])


def result_cf():
    """Result table of cashflows

    .. seealso::

       * :func:`premiums`
       * :func:`claims`
       * :func:`expenses`
       * :func:`commissions`
       * :func:`net_cf`

    """

    t_len = range(proj_len())

    data = {
        "Premiums": [premiums(t) for t in t_len],
        "Claims": [claims(t) for t in t_len],
        "Expenses": [expenses(t) for t in t_len],
        "Commissions": [commissions(t) for t in t_len],
        "Net Cashflow": [net_cf(t) for t in t_len]
    }
    return pd.DataFrame.from_dict(data)


def result_pols():
    """Result table of policy decrement

    .. seealso::

       * :func:`pols_if`
       * :func:`pols_maturity`
       * :func:`pols_new_biz`
       * :func:`pols_death`
       * :func:`pols_lapse`

    """

    t_len = range(proj_len())

    data = {
        "pols_if": [pols_if(t) for t in t_len],
        "pols_maturity": [pols_maturity(t) for t in t_len],
        "pols_new_biz": [pols_new_biz(t) for t in t_len],
        "pols_death": [pols_death(t) for t in t_len],
        "pols_lapse": [pols_lapse(t) for t in t_len]
    }

    return pd.DataFrame.from_dict(data)


def result_pv():
    """Result table of present value of cashflows

    .. seealso::

       * :func:`pv_premiums`
       * :func:`pv_claims`
       * :func:`pv_expenses`
       * :func:`pv_commissions`
       * :func:`pv_net_cf`

    """

    cols = ["Premiums", "Claims", "Expenses", "Commissions", "Net Cashflow"]
    pvs = [pv_premiums(), pv_claims(), pv_expenses(), pv_commissions(), pv_net_cf()]

    return pd.DataFrame.from_dict(
            data={"PV": pvs},
            columns=cols,
            orient='index')


def sex(): 
    """The sex of the selected model point

    .. note::
       This cells is not used by default.

    The element labeled ``sex`` of the Series returned by
    :func:`model_point`.
    """
    return model_point()["sex"]


def sum_assured():
    """The sum assured of the selected model point

    The element labeled ``sum_assured`` of the Series returned by
    :func:`model_point`.
    """
    return model_point()["sum_assured"]


# ---------------------------------------------------------------------------
# References

disc_rate_ann = ("DataClient", 2160330731208)

model_point_table = ("DataClient", 2160335976392)

mort_table = ("DataClient", 2160317735752)

np = ("Module", "numpy")

pd = ("Module", "pandas")

point_id = 1

premium_table = ("DataClient", 2160336367816)


r"""
Black-Scholes-Merton on dividend paying stock
=============================================

As the :doc:`/libraries/notebooks/savings/savings_example2` shows,
time values of options and guarantees on a GMAB policy
can be calculated using the Black-Scholes-Merton formula
on a dividend paying stock, when maintenance fees are deducted
from account value at a constant rate, by regarding the fees as dividends.

The Black-Scholes-Merton pricing formula for European put options
on a dividend paying stock
can be expressed as below, where
:math:`X`, :math:`S_{0}`, :math:`q` correspond to the sum assured,
the initial account value and the maintenence fee rate(1%) in this example.

.. math::

    p=Xe^{-rT}N\left(-d_{2}\right)-S_{0}e^{-qT}N\left(-d_{1}\right)

    d_{1}=d_{1}=\frac{\ln\left(\frac{S_{0}}{X}\right)+\left(r-q+\frac{\sigma^{2}}{2}\right)T}{\sigma\sqrt{T}}

    d_{2}=d_{1}-\sigma\sqrt{T}

The graph below compares the option values with the maintenance fee deduction
against the corresponding values without fee deduction
for various in-the-moneyness.

Reference: *Options, Futures, and Other Derivatives* by John C.Hull

.. seealso::

    * :doc:`/libraries/notebooks/savings/savings_example1` notebook in the :mod:`~savings` library
    * :doc:`/libraries/notebooks/savings/savings_example2` notebook in the :mod:`~savings` library


"""
import modelx as mx
import pandas as pd
import matplotlib.pyplot as plt
from scipy.stats import norm, lognorm
import numpy as np

ex1 = mx.read_model("CashValue_ME_EX1").Projection
ex2 = mx.read_model("CashValue_ME_EX2").Projection

ex1.model_point_table = ex1.model_point_moneyness
ex2.model_point_table = ex2.model_point_moneyness
S0 = ex1.model_point_table['premium_pp'] * ex1.model_point_table['policy_count']

fig, ax = plt.subplots()
ax.scatter(S0, ex1.formula_option_put(120), s= 10, alpha=1, label='No dividends')
ax.scatter(S0, ex2.formula_option_put(120), alpha=0.5, label='With dividends')
ax.legend()
ax.grid(True)
fig.suptitle('TVOG by ITM')

ex1.model.close()
ex2.model.close()







from modelx.serialize.jsonvalues import *

_name = "solvency2"

_allow_none = False

_spaces = [
    "Input",
    "LifeTable",
    "Economic",
    "BaseProj",
    "PV",
    "Override",
    "SCR_life",
    "Policy",
    "Assumptions"
]




"""The main Space in the :mod:`~basiclife.BasicTerm_M` model.

:mod:`~basiclife.BasicTerm_M.Projection` is the only Space defined
in the :mod:`~basiclife.BasicTerm_M` model, and it contains
all the logic and data used in the model.

.. rubric:: Parameters and References

(In all the sample code below,
the global variable ``Projection`` refers to the
:mod:`~basiclife.BasicTerm_M.Projection` Space.)

Attributes:

    model_point_table: All model point data as a DataFrame.
        The sample model point data was generated by
        *generate_model_points.ipynb* included in the library.
        By default, :func:`model_point` returns this :attr:`model_point_table`.
        The DataFrame has columns labeled ``age_at_entry``,
        ``sex``, ``policy_term``, ``policy_count``
        and ``sum_assured``.
        Cells defined in :mod:`~basiclife.BasicTerm_M.Projection`
        with the same names as these columns return
        the corresponding columns.
        (``policy_count`` is not used by default.)

        .. code-block::

            >>> Projection.model_poit_table
                       age_at_entry sex  policy_term  policy_count  sum_assured
            point_id
            1                    47   M           10             1       622000
            2                    29   M           20             1       752000
            3                    51   F           10             1       799000
            4                    32   F           20             1       422000
            5                    28   M           15             1       605000
                            ...  ..          ...           ...          ...
            9996                 47   M           20             1       827000
            9997                 30   M           15             1       826000
            9998                 45   F           20             1       783000
            9999                 39   M           20             1       302000
            10000                22   F           15             1       576000

            [10000 rows x 5 columns]

        The DataFrame is saved in the Excel file *model_point_table.xlsx*
        placed in the model folder.
        :attr:`model_point_table` is created by
        Projection's `new_pandas`_ method,
        so that the DataFrame is saved in the separate file.
        The DataFrame has the injected attribute
        of ``_mx_dataclident``::

            >>> Projection.model_point_table._mx_dataclient
            <PandasData path='model_point_table.xlsx' filetype='excel'>

        .. seealso::

           * :func:`model_point`
           * :func:`age_at_entry`
           * :func:`sex`
           * :func:`policy_term`
           * :func:`sum_assured`


    disc_rate_ann: Annual discount rates by duration as a pandas Series.

        .. code-block::

            >>> Projection.disc_rate_ann
            year
            0      0.00000
            1      0.00555
            2      0.00684
            3      0.00788
            4      0.00866

            146    0.03025
            147    0.03033
            148    0.03041
            149    0.03049
            150    0.03056
            Name: disc_rate_ann, Length: 151, dtype: float64

        The Series is saved in the Excel file *disc_rate_ann.xlsx*
        placed in the model folder.
        :attr:`disc_rate_ann` is created by
        Projection's `new_pandas`_ method,
        so that the Series is saved in the separate file.
        The Series has the injected attribute
        of ``_mx_dataclident``::

            >>> Projection.disc_rate_ann._mx_dataclient
            <PandasData path='disc_rate_ann.xlsx' filetype='excel'>

        .. seealso::

           * :func:`disc_rate_mth`
           * :func:`disc_factors`

    mort_table: Mortality table by age and duration as a DataFrame.
        See *basic_term_sample.xlsx* included in this library
        for how the sample mortality rates are created.

        .. code-block::

            >>> Projection.mort_table
                        0         1         2         3         4         5
            Age
            18   0.000231  0.000254  0.000280  0.000308  0.000338  0.000372
            19   0.000235  0.000259  0.000285  0.000313  0.000345  0.000379
            20   0.000240  0.000264  0.000290  0.000319  0.000351  0.000386
            21   0.000245  0.000269  0.000296  0.000326  0.000359  0.000394
            22   0.000250  0.000275  0.000303  0.000333  0.000367  0.000403
            ..        ...       ...       ...       ...       ...       ...
            116  1.000000  1.000000  1.000000  1.000000  1.000000  1.000000
            117  1.000000  1.000000  1.000000  1.000000  1.000000  1.000000
            118  1.000000  1.000000  1.000000  1.000000  1.000000  1.000000
            119  1.000000  1.000000  1.000000  1.000000  1.000000  1.000000
            120  1.000000  1.000000  1.000000  1.000000  1.000000  1.000000

            [103 rows x 6 columns]

        The DataFrame is saved in the Excel file *mort_table.xlsx*
        placed in the model folder.
        :attr:`mort_table` is created by
        Projection's `new_pandas`_ method,
        so that the DataFrame is saved in the separate file.
        The DataFrame has the injected attribute
        of ``_mx_dataclident``::

            >>> Projection.mort_table._mx_dataclient
            <PandasData path='mort_table.xlsx' filetype='excel'>

        .. seealso::

           * :func:`mort_rate`
           * :func:`mort_rate_mth`

    np: The `numpy`_ module.
    pd: The `pandas`_ module.

.. _numpy:
   https://numpy.org/

.. _pandas:
   https://pandas.pydata.org/

.. _new_pandas:
   https://docs.modelx.io/en/latest/reference/space/generated/modelx.core.space.UserSpace.new_pandas.html

"""

from modelx.serialize.jsonvalues import *

_formula = None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def age(t):
    """The attained age at time t.

    Defined as::

        age_at_entry() + duration(t)

    .. seealso::

        * :func:`age_at_entry`
        * :func:`duration`

    """
    return age_at_entry() + duration(t)


def age_at_entry():
    """The age at entry of the model points

    The ``age_at_entry`` column of the DataFrame returned by
    :func:`model_point`.
    """
    return model_point()["age_at_entry"]


def claim_pp(t):
    """Claim per policy

    The claim amount per plicy. Defaults to :func:`sum_assured`.
    """
    return sum_assured()


def claims(t):
    """Claims

    Claims during the period from ``t`` to ``t+1`` defined as::

        claim_pp(t) * pols_death(t)

    .. seealso::

        * :func:`claim_pp`
        * :func:`pols_death`

    """
    return claim_pp(t) * pols_death(t)


def commissions(t):
    """Commissions

    By default, 100% premiums for the first year, 0 otherwise.

    .. seealso::

        * :func:`premiums`
        * :func:`duration`

    """
    return (duration(t) == 0) * premiums(t)


def disc_factors():
    """Discount factors.

    Vector of the discount factors as a Numpy array. Used for calculating
    the present values of cashflows.

    .. seealso::

        :func:`disc_rate_mth`
    """
    return np.array(list((1 + disc_rate_mth()[t])**(-t) for t in range(max_proj_len())))


def disc_rate_mth():
    """Monthly discount rate

    Nummpy array of monthly discount rates from time 0 to :func:`max_proj_len` - 1
    defined as::

        (1 + disc_rate_ann)**(1/12) - 1

    .. seealso::

        :func:`disc_rate_ann`

    """
    return np.array(list((1 + disc_rate_ann[t//12])**(1/12) - 1 for t in range(max_proj_len())))


def duration(t):
    """Duration in force in years"""
    return t//12


def expense_acq():
    """Acquisition expense per policy

    ``300`` by default.
    """
    return 300


def expense_maint():
    """Annual maintenance expense per policy

    ``60`` by default.
    """
    return 60


def expenses(t):
    """Acquisition and maintenance expenses

    Expense cashflow during the period from ``t`` to ``t+1``.
    For any ``t``, the maintenance expense is recognized,
    which is defined as::

        pols_if(t) * expense_maint()/12 * inflation_factor(t)

    At ``t=0`` only, the acquisition expense,
    defined as :func:`expense_acq`, is recognized.

    .. seealso::

        * :func:`pols_if`
        * :func:`expense_maint`
        * :func:`inflation_factor`

    .. versionchanged:: 0.2.0
       The maintenance expense is also recognized for ``t=0``.

    """
    return (t == 0) * expense_acq() * pols_if(t) \
           + pols_if(t) * expense_maint()/12 * inflation_factor(t)


def inflation_factor(t):
    """The inflation factor at time t

    .. seealso::

        * :func:`inflation_rate`

    """
    return (1 + inflation_rate())**(t/12)


def inflation_rate():
    """Inflation rate"""
    return 0.01


def lapse_rate(t):
    """Lapse rate

    By default, the lapse rate assumption is defined by duration as::

        max(0.1 - 0.02 * duration(t), 0.02)

    .. seealso::

        :func:`duration`

    """
    return np.maximum(0.1 - 0.02 * duration(t), 0.02)


def loading_prem():
    """Loading per premium

    ``0.5`` by default.

    .. seealso::

        * :func:`premium_pp`

    """
    return 0.5


max_proj_len = lambda: max(proj_len())
"""The max of all projection lengths

Defined as ``max(proj_len())``

.. seealso::
    :func:`proj_len`
"""

def model_point():
    """Target model points

    Returns as a DataFrame the model points to be in the scope of calculation.
    By default, this Cells returns the entire :attr:`model_point_table`
    without change.
    To select model points, change this formula so that this
    Cells returns a DataFrame that contains only the selected model points.

    Examples:
        To select only the model point 1::

            def model_point():
                return model_point_table.loc[1:1]

        To select model points whose ages at entry are 40 or greater::

            def model_point():
                return model_point_table[model_point_table["age_at_entry"] >= 40]

    Note that the shape of the returned DataFrame must be the
    same as the original DataFrame, i.e. :attr:`model_point_table`.

    When selecting only one model point, make sure the
    returned object is a DataFrame, not a Series, as seen in the example
    above where ``model_point_table.loc[1:1]`` is specified
    instead of ``model_point_table.loc[1]``.

    Be careful not to accidentally change the original table.
    """
    return model_point_table


def mort_rate(t):
    """Mortality rate to be applied at time t

    .. seealso::

       * :attr:`mort_table`
       * :func:`mort_rate_mth`

    """
    result = mort_table[str(min(5, duration(t)))][age(t)]
    result.index = model_point().index
    return result


def mort_rate_mth(t):
    """Monthly mortality rate to be applied at time t

    .. seealso::

       * :attr:`mort_table`
       * :func:`mort_rate`

    """
    return 1-(1- mort_rate(t))**(1/12)


def net_cf(t):
    """Net cashflow

    Net cashflow for the period from ``t`` to ``t+1`` defined as::

        premiums(t) - claims(t) - expenses(t) - commissions(t)

    .. seealso::

        * :func:`premiums`
        * :func:`claims`
        * :func:`expenses`
        * :func:`commissions`

    """
    return premiums(t) - claims(t) - expenses(t) - commissions(t)


def net_premium_pp():
    """Net premium per policy

    The net premium per policy is defined so that
    the present value of net premiums equates to the present value of
    claims::

        pv_claims() / pv_pols_if()

    .. seealso::

        * :func:`pv_claims`
        * :func:`pv_pols_if`

    """
    return pv_claims() / pv_pols_if()


def policy_term():
    """The policy term of the model points.

    The ``policy_term`` column of the DataFrame returned by
    :func:`model_point`.
    """
    return model_point()["policy_term"]


def pols_death(t):
    """Number of death occurring at time t"""
    return pols_if(t) * mort_rate_mth(t)


def pols_if(t):
    """Number of policies in-force

    Number of in-force policies calculated recursively.
    The initial value is read from :func:`pols_if_init`.
    Subsequent values are defined recursively as::

        pols_if(t-1) - pols_lapse(t-1) - pols_death(t-1) - pols_maturity(t)

    .. seealso::
        * :func:`pols_lapse`
        * :func:`pols_death`
        * :func:`pols_maturity`

    """
    if t==0:
        return pols_if_init()
    elif t < max_proj_len():
        return pols_if(t-1) - pols_lapse(t-1) - pols_death(t-1) - pols_maturity(t)
    else:
        raise KeyError("t out of range")


def pols_if_init():
    """Initial Number of Policies In-force

    Number of in-force policies at time 0 referenced from :func:`pols_if`.
    Defaults to 1.
    """
    return pd.Series(1, index=model_point().index)


def pols_lapse(t):
    """Number of lapse occurring at time t

    .. seealso::
        * :func:`pols_if`
        * :func:`lapse_rate`

    """
    return (pols_if(t) - pols_death(t)) * (1-(1 - lapse_rate(t))**(1/12))


def pols_maturity(t):
    """Number of maturing policies

    The policy maturity occurs at ``t == 12 * policy_term()``,
    after death and lapse during the last period::

        pols_if(t-1) - pols_lapse(t-1) - pols_death(t-1)

    otherwise ``0``.
    """
    return (t == policy_term() * 12) * (pols_if(t-1) - pols_lapse(t-1) - pols_death(t-1))


def premium_pp():
    """Monthly premium per policy

    Monthly premium amount per policy defined as::

        round((1 + loading_prem()) * net_premium(), 2)

    .. versionchanged:: 0.2.0
       The ``t`` parameter is removed.

    .. seealso::

        * :func:`loading_prem`
        * :func:`net_premium_pp`

    """
    return np.around((1 + loading_prem()) * net_premium_pp(), 2)


def premiums(t):
    """Premium income

    Premium income during the period from ``t`` to ``t+1`` defined as::

        premium_pp(t) * pols_if(t)

    .. seealso::

        * :func:`premium_pp`
        * :func:`pols_if`

    """
    return premium_pp() * pols_if(t)


def proj_len():
    """Projection length in months

    Projection length in months defined as::

        12 * policy_term() + 1

    Since this model carries out projections for all the model points
    simultaneously, the projections are actually carried out
    from 0 to :attr:`max_proj_len` for all the model points.

    .. seealso::

        :func:`policy_term`

    """
    return 12 * policy_term() + 1


def pv_claims():
    """Present value of claims

    .. seealso::

        * :func:`claims`

    """
    cl = np.array(list(claims(t) for t in range(max_proj_len()))).transpose()

    return cl @ disc_factors()[:max_proj_len()]


def pv_commissions():
    """Present value of commissions

    .. seealso::

        * :func:`expenses`

    """
    result = np.array(list(commissions(t) for t in range(max_proj_len()))).transpose()

    return result @ disc_factors()[:max_proj_len()]


def pv_expenses():
    """Present value of expenses

    .. seealso::

        * :func:`expenses`

    """
    result = np.array(list(expenses(t) for t in range(max_proj_len()))).transpose()

    return result @ disc_factors()[:max_proj_len()]


def pv_net_cf():
    """Present value of net cashflows.

    Defined as::

        pv_premiums() - pv_claims() - pv_expenses() - pv_commissions()

    .. seealso::

        * :func:`pv_premiums`
        * :func:`pv_claims`
        * :func:`pv_expenses`
        * :func:`pv_commissions`

    """
    return pv_premiums() - pv_claims() - pv_expenses() - pv_commissions()


def pv_pols_if():
    """Present value of policies in-force

    The discounted sum of the number of in-force policies at each month.
    It is used as the annuity factor for calculating :func:`net_premium_pp`.

    """
    result = np.array(list(pols_if(t) for t in range(max_proj_len()))).transpose()

    return result @ disc_factors()[:max_proj_len()]


def pv_premiums():
    """Present value of premiums

    .. seealso::

        * :func:`premiums`

    """
    result = np.array(list(premiums(t) for t in range(max_proj_len()))).transpose()

    return result @ disc_factors()[:max_proj_len()]


def result_cf():
    """Result table of cashflows

    .. seealso::

       * :func:`premiums`
       * :func:`claims`
       * :func:`expenses`
       * :func:`commissions`
       * :func:`net_cf`

    """

    t_len = range(max_proj_len())

    data = {
        "Premiums": [sum(premiums(t)) for t in t_len],
        "Claims": [sum(claims(t)) for t in t_len],
        "Expenses": [sum(expenses(t)) for t in t_len],
        "Commissions": [sum(commissions(t)) for t in t_len],
        "Net Cashflow": [sum(net_cf(t)) for t in t_len]
    }

    return pd.DataFrame(data, index=t_len)


def result_pv():
    """Result table of present value of cashflows

    .. seealso::

       * :func:`pv_premiums`
       * :func:`pv_claims`
       * :func:`pv_expenses`
       * :func:`pv_commissions`
       * :func:`pv_net_cf`

    """


    data = {
        "PV Premiums": pv_premiums(),
        "PV Claims": pv_claims(),
        "PV Expenses": pv_expenses(),
        "PV Commissions": pv_commissions(),
        "PV Net Cashflow": pv_net_cf()
    }

    return pd.DataFrame(data, index=model_point().index)


def sex():
    """The sex of the model points

    .. note::
       This cells is not used by default.

    The ``sex`` column of the DataFrame returned by
    :func:`model_point`.
    """
    return model_point()["sex"]


def sum_assured():
    """The sum assured of the model points

    The ``sum_assured`` column of the DataFrame returned by
    :func:`model_point`.
    """
    return model_point()["sum_assured"]


# ---------------------------------------------------------------------------
# References

disc_rate_ann = ("DataClient", 2105121422280)

model_point_table = ("DataClient", 2105121417096)

mort_table = ("DataClient", 2105119667144)

np = ("Module", "numpy")

pd = ("Module", "pandas")


"""Commutation functions and actuarial notations

The ``LifeTable`` Space provides
commutation functions and actuarial notations, such as
:math:`D_{x}` and :math:`\\require{enclose}{}_{f|}\\overline{A}_{x}`.
Mortality tables are read from *input.xlsx* into an `ExcelRange`_ object.
The `ExcelRange`_ object is bound to a Reference, :attr:`MortalityTable`.

This Space is included in:

* :mod:`simplelife`
* :mod:`nestedlife`
* :mod:`ifrs17sim`
* :mod:`solvency2`

.. _ExcelRange:
   https://docs.modelx.io/en/latest/reference/dataclient.html#excelrange

.. rubric:: Parameters

``LifeTable`` Space is parameterized with :attr:`Sex`,
:attr:`IntRate` and :attr:`TableID`::

        >>> simplelife.LifeTable.parameters
        ('Sex', 'IntRate', 'TableID')

Each ItemSpace represents commutations functions actuarial notations
for a combination of :attr:`Sex`, :attr:`IntRate` and :attr:`TableID`.
For example, ``LifeTable['M', 0.03, 1]`` contains commutation functions
and actuarial notations for Male, the interest rate of 3%, mortality table 1.


Attributes:
    Sex(:obj:`str`): 'M' or 'F' to indicate male or female column in the mortality table.
    IntRate(:obj:`float`): The constant interest rate for discounting.
    TableID(:obj:`int`): The identifier of the mortality table


.. rubric:: References

Attributes:
    MortalityTable: `ExcelRange`_ object holding mortality tables.
        The data is read from *MortalityTables* range in *input.xlsx*.

Example:

    An example of ``LifeTable`` in the :mod:`simplelife` model::

        >>> simplelife.LifeTable['M', 0.03, 1].AnnDuenx(40, 10)
        8.725179890621531

External Links:
    * `International actuarial notation by F.S.Perryman <https://www.casact.org/pubs/proceed/proceed49/49123.pdf>`_
    * `Actuarial notations on Wikipedia <https://en.wikipedia.org/wiki/Actuarial_notation>`_


"""

from modelx.serialize.jsonvalues import *

_formula = lambda Sex, IntRate, TableID: None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def AnnDuenx(x, n, k=1, f=0):
    """ The present value of an annuity-due.

    .. math::

        \\require{enclose}{}_{f|}\\ddot{a}_{x:\\enclose{actuarial}{n}}^{(k)}


    Args:
        x(int): age
        n(int): length of annuity payments in years
        k(int, optional): number of split payments in a year
        f(int, optional): waiting period in years

    """
    if Dx(x) == 0:
        return 0

    result = (Nx(x+f) - Nx(x+f+n)) / Dx(x)

    if k > 1:
        return result - (k-1) / (2*k) * (1 - Dx(x+f+n) / Dx(x))
    else:
        return result


def AnnDuex(x, k, f=0):
    """The present value of a lifetime annuity due.

    Args:
        x(int): age
        k(int, optional): number of split payments in a year
        f(int, optional): waiting period in years
    """
    if Dx(x) == 0:
        return 0

    result = (Nx(x+f)) / Dx(x)

    if k > 1:
        return result - (k-1) / (2*k)
    else:
        return result


def Ax(x, f=0):
    """The present value of a lifetime assurance on a person at age ``x``
    payable immediately upon death, optionally with an waiting period of ``f`` years.

    .. math::

        \\require{enclose}{}_{f|}\\overline{A}_{x}
    """
    if Dx(x) == 0:
        return 0
    else:
        return Mx(x+f) / Dx(x)


def Axn(x, n, f=0):
    """The present value of an assurance on a person at age ``x`` payable
    immediately upon death, optionally with an waiting period of ``f`` years.

    .. math::

        \\require{enclose}{}_{f|}\\overline{A}^{1}_{x:\\enclose{actuarial}{n}}

    """
    if Dx(x) == 0:
        return 0
    else:
        return (Mx(x+f) - Mx(x+f+n)) / Dx(x)


def Cx(x):
    """The commutation column :math:`\\overline{C_x}`.
    """

    return dx(x) * disc**(x+1/2)


def Dx(x):
    """The commutation column :math:`D_{x} = l_{x}v^{x}`.
    """
    return lx(x) * disc ** x


def Exn(x, n):
    """ The value of an endowment on a person at age ``x``
    payable after n years

    .. math::

        {}_{n}E_x

    """
    if Dx(x) == 0:
        return 0
    else:
        return Dx(x+n) / Dx(x)


def Mx(x):
    """The commutation column :math:`M_x`."""

    if x >= 110:
        return Dx(x)
    else:
        return Mx(x+1) + Cx(x)


def Nx(x):
    """The commutation column :math:`N_x`."""
    if x >= 110:    # TODO: Get the last age from the table
        return Dx(x)
    else:
        return Nx(x+1) + Dx(x)


def disc():
    """The discount factor :math:`v = 1/(1 + i)`."""
    return 1 / (1 + IntRate)


def dx(x):
    """The number of persons who die between ages ``x`` and ``x+1``"""
    return lx(x) * qx(x)


def lx(x):
    """The number of persons remaining at age ``x``. """
    if x == 0:
        return 100000
    else:
        return lx(x-1) - dx(x-1)


def qx(x):
    """Probability that a person at age ``x`` will die in one year."""
    return MortalityTable[TableID, Sex, x]


# ---------------------------------------------------------------------------
# References

Sex = "M"

IntRate = 0.01

TableID = 1

MortalityTable = ("Pickle", 2325070426248)


"""Override module for the expense risk calculation

The formulas in this module add or overrides cells related to expense in
:mod:`projection <solvency2.projection>` module.
"""

from modelx.serialize.jsonvalues import *

_formula = None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def InflFactor(t):
    """Inflation factor reflecting expense shocks"""
    if t == 0:
        return 1
    else:        
        if t >= t0:
            shock = Factor(Risk, Shock, Scope, 'inflation')
        else:
            shock = 0

        return InflFactor(t-1) * (1 + asmp.InflRate + shock)


def SizeExpsAcq(t):
    """Acquisition expense per policy at time t"""
    if t == 0:
        return (SizeAnnPrem(t) * asmp.ExpsAcqAnnPrem
                + (SizeSumAssured(t) * asmp.ExpsAcqSA + asmp.ExpsAcqPol)
                * InflFactor(t) / InflFactor(0))
    else:
        return 0


def SizeExpsMaint(t):
    """Maintenance expense per policy at time t"""

    shock = Factor(Risk, Shock, Scope)

    return (SizeAnnPrem(t) * asmp.ExpsMaintAnnPrem
            + (SizeSumAssured(t) * asmp.ExpsMaintSA + asmp.ExpsMaintPol)
            * InflFactor(t)) * (1 + shock)





"""Present Value mix-in Space

This Space serves as a base Space for :mod:`~simplelife.model.Projection`
Space, and it contains Cells to take the present value of projected cashflows.

.. blockdiag::

   blockdiag {
     default_node_color="#D5E8D4";
     default_linecolor="#628E47";
     BaseProj[style=dotted]
     BaseProj <- Projection [hstyle=generalization]
     PV[style=dotted]
     PV <- Projection [hstyle=generalization];
   }

"""

from modelx.serialize.jsonvalues import *

_formula = None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def InterestNetCF(t):
    """Interest accreted on pv of net cashflows"""
    if t > last_t:
        return 0
    else:
        return (PV_NetCashflow(t)
                - PremIncome(t)
                + ExpsTotal(t)) * DiscRate(t)


def PV_BenefitDeath(t):
    """Present value of death benefits"""
    if t > last_t:
        return 0
    else:
        return (-BenefitDeath(t) + PV_BenefitDeath(t+1)) / (1 + DiscRate(t))


def PV_BenefitMat(t):
    """Present value of matuirty benefits"""
    if t > last_t:
        return 0
    else:
        return (-BenefitMat(t) + PV_BenefitMat(t+1)) / (1 + DiscRate(t))


def PV_BenefitSurr(t):
    """Present value of surrender benefits"""
    if t > last_t:
        return 0
    else:
        return (-BenefitSurr(t) + PV_BenefitSurr(t+1)) / (1 + DiscRate(t))


def PV_BenefitTotal(t):
    """Present value of total benefits"""
    if t > last_t:
        return 0
    else:
        return (-BenefitTotal(t) + PV_BenefitTotal(t+1)) / (1 + DiscRate(t))


def PV_Check(t):
    return PV_NetCashflow(t) - PV_NetCashflowForCheck(t)


def PV_ExpsAcq(t):
    """Present value of acquisition expenses"""
    if t > last_t:
        return 0
    else:
        return - ExpsAcq(t) + PV_ExpsAcq(t+1) / (1 + DiscRate(t))


def PV_ExpsCommTotal(t):
    """Present value of commission expenses"""
    if t > last_t:
        return 0
    else:
        return - ExpsCommTotal(t) + PV_ExpsCommTotal(t+1) / (1 + DiscRate(t))


def PV_ExpsMaint(t):
    """Present value of maintenance expenses"""
    if t > last_t:
        return 0
    else:
        return - ExpsMaint(t) + PV_ExpsMaint(t+1) / (1 + DiscRate(t))


def PV_ExpsTotal(t):
    """Present value of total expenses"""
    if t > last_t:
        return 0
    else:
        return - ExpsTotal(t) + PV_ExpsTotal(t+1) / (1 + DiscRate(t))


def PV_NetCashflow(t):
    """Present value of net cashflow"""
    return (PV_PremIncome(t)
            + PV_ExpsTotal(t)
            + PV_BenefitTotal(t))


def PV_NetCashflowForCheck(t):
    """Present value of net cashflow"""
    if t > last_t:
        return 0
    else:
        return (PremIncome(t)
                - ExpsTotal(t)
                - BenefitTotal(t) / (1 + DiscRate(t))
                + PV_NetCashflow(t+1) / (1 + DiscRate(t)))


def PV_PremIncome(t):
    """Present value of premium income"""
    if t > last_t:
        return 0
    else:
        return PremIncome(t) + PV_PremIncome(t+1) / (1 + DiscRate(t))


def PV_SumInsurIF(t):
    """Present value of insurance in-force"""
    if t > last_t:
        return 0
    else:
        return InsurIF_Beg1(t) + PV_SumInsurIF(t+1) / (1 + DiscRate(t))





"""Economic Assumptions

This Space includes economic assumptions, such as discount rates
and inflation rates. This Space is parametrized with :attr:`ScenID`.

.. rubric:: Space Parameters

Attributes:
    ScenID(:obj:`int`): Scenario ID

.. rubric:: References

Attributes:
    AsmpLookup: Reference to :func:`fastlife.model.Projection.Assumptions.AsmpLookup`
    Scenarios: `ExcelRange`_ object holding scenario data read from
        the input file.

.. _ExcelRange:
   https://docs.modelx.io/en/latest/reference/dataclient.html#excelrange

"""

from modelx.serialize.jsonvalues import *

_formula = lambda ScenID: None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def DiscRate(t):
    """Rates for discount cashflows"""
    return Scenarios[ScenID, "IntRate", t]


def InflFactor(t):
    """Inflation factors to adjust expense cashflows"""
    if t == 0:
        return 1
    else:
        return InflFactor(t-1) / (1 + AsmpLookup("InflRate"))


def InvstRetRate(t):
    """Rate of investment return

    Set equal to the :func:`DiscRate`
    """
    return DiscRate(t)


# ---------------------------------------------------------------------------
# References

ScenID = 1

Scenarios = ("Pickle", 3020541690632)

AsmpLookup = ("Interface", ("..", "Projection", "Assumptions", "AsmpLookup"), "auto")


"""
Smith-Wilson extrapolation with various alphas
==============================================

This notebook reads in the **smithwilson** model,
and extrapolates forward rates
using the Smith-Wilson method with various alphas, and draws
the forward rates.
See :doc:`/projects/smithwilson` for the details of the **smithwilson** model.

The source of the input data is Switzerland EIOPA spot rates
with LLP 25 years available from the following source.

Source: RFR_spot_no_VA tab in EIOPA_RFR_20190531_Term_Structures.xlsx,
archived in EIOPA_RFR_20190531.zip, avaialble on
`EIOPA's Risk-Free Interest Rate Term Structures web site`_.

.. _EIOPA's Risk-Free Interest Rate Term Structures web site: https://wayback.archive-it.org/org-1495/20191229100044/https:/eiopa.europa.eu/regulation-supervision/insurance/solvency-ii-technical-information/risk-free-interest-rate-term-structures


"""

import modelx as mx
import pandas as pd

#%%
# The code below reads the ``smithwilson`` model from *"model"* folder,
# and assign ``SmithWilson`` space in the model to
# a global variable named ``space``.

space = mx.read_model("model").SmithWilson

#%%
# The code below is for parametrizing  the ``SmithWilson`` space by ``x``.
# The parameter ``x`` takes integer values, and they can be positive and negative.
# the alpha in the ``SmithWilson[x]`` space is adjusted from the original alpha
# by ``x`` times 10%.

def parametrize(x):
    return {"refs": {"alpha": alpha * (1 + 0.1*x)}}

space.formula = parametrize

#%%
# The code below creates the ``ForwardRates`` cells,
# which calculates extrapolated forward rates from
# the present values of bond prices.

@mx.defcells
def ForwardRates(t):
    return P(t) / P(t+1) - 1


#%%
# The code below plots the extrapolated forward rates
# with various alphas.

d = {"{0:.3}".format(space.alpha * (1 + 0.1*x)):
         [space[x].ForwardRates[t] for t in range(1, 101)]
     for x in range(-6, 5)}

df = pd.DataFrame(d, index=range(1, 101))

df.plot.line()


"""Assumption input and calculations for individual policies.

This Space is a child Space of :mod:`~simplelife.model.Projection`,
and it holds assumption parameters and rates used
by the :mod:`~simplelife.model.Projection` Space.

.. blockdiag::

   blockdiag {
     default_node_color="#D5E8D4";
     default_linecolor="#628E47";
     node_width=150;
     Proj[label="Projection\\n[PolicyID, ScenID=1]", stacked];
     Proj <- Assumptions [hstyle=composition];
     Proj <- Policy [hstyle=composition];
   }


.. rubric:: Parameters

Since :mod:`~simplelife.model.Projection` is parameterized with
:attr:`PolicyID` and :attr:`ScenID`, this Space is also
parameterized as a child space of :mod:`~simplelife.model.Projection`.
For example, ``simplelife.Projection[1].Assumptions.ExpsMaintSA()``
represents the expense maintenance per sum assured for Policy 1.

Attributes:
    PolicyID(:obj:`int`): Policy ID
    ScenID(:obj:`int`, optional): Scenario ID, defaults to 1.

.. rubric:: References

Attributes:
    AssumptionTables: `ExcelRange`_ object holding data read from the
        Excel range *AssumptionTable* in *input.xlsx*.

    MortalityTable: `ExcelRange`_ object holding mortality tables.
        The data is read from *MortalityTables* range in *input.xlsx*.

    prod: Alias for :func:`simplelife.Projection.Policy.Product`
    polt: Alias for :func:`simplelife.Projection.Policy.PolicyType`
    gen: Alias for :func:`simplelife.Projection.Policy.Gen`
    sex: Alias for :func:`simplelife.Projection.Policy.Sex`
    AsmpLookup: Alias for :func:`simplelife.Input.AsmpLookup`

.. _ExcelRange:
   https://docs.modelx.io/en/latest/reference/dataclient.html#excelrange

"""

from modelx.serialize.jsonvalues import *

_formula = None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def BaseMortRate(x):
    """Bae mortality rate"""

    table_id = AsmpLookup.match("BaseMort", prod(), polt(), gen()).value
    return MortalityTables[table_id, sex(), x]


def CnsmpTax():
    """Consumption tax rate"""
    return AsmpLookup("CnsmpTax")


def CommInitPrem():
    """Initial commission per premium"""
    result = AsmpLookup.match("CommInitPrem", prod(), polt(), gen()).value

    if result is not None:
        return result
    else:
        raise ValueError('CommInitPrem not found')


def CommRenPrem():
    """Renewal commission per premium"""
    result = AsmpLookup.match("CommRenPrem", prod(), polt(), gen()).value

    if result is not None:
        return  result
    else:
        raise ValueError('CommRenPrem not found')


def CommRenTerm():
    """Renewal commission term"""
    result = AsmpLookup.match("CommRenTerm", prod(), polt(), gen()).value

    if result is not None:
        return result
    else:
        raise ValueError('CommRenTerm not found')


def ExpsAcqAnnPrem():
    """Acquisition expense per annualized premium"""
    return AsmpLookup.match("ExpsAcqAnnPrem", prod(), polt(), gen()).value


def ExpsAcqPol():
    """Acquisition expense per policy"""
    return AsmpLookup.match("ExpsAcqPol", prod(), polt(), gen()).value


def ExpsAcqSA():
    """Acquisition expense per sum assured"""
    return AsmpLookup.match("ExpsAcqSA", prod(), polt(), gen()).value


def ExpsMaintAnnPrem():
    """Maintenance expense per annualized premium"""
    return AsmpLookup.match("ExpsMaintPrem", prod(), polt(), gen()).value


def ExpsMaintPol():
    """Maintenance expense per policy"""
    return AsmpLookup.match("ExpsMaintPol", prod(), polt(), gen()).value


def ExpsMaintSA():
    """Maintenance expense per sum assured"""
    return AsmpLookup.match("ExpsMaintSA", prod(), polt(), gen()).value


def InflRate():
    """Inflation rate"""
    return AsmpLookup("InflRate")


def LastAge():
    """Age at which mortality becomes 1"""
    x = 0
    while True:
        if BaseMortRate(x) == 1:
            return x
        x += 1


def MortFactor(y):
    """Mortality factor"""
    table = AsmpLookup.match("MortFactor", prod(), polt(), gen()).value

    if table is None:
        raise ValueError('MortFactor not found')

    result = AssumptionTables.get((table, y), None)

    if result is None:
        return MortFactor(y-1)
    else:
        return result


def MortTable():
    """Mortality Table"""
    result = AsmpLookup.match("BaseMort", prod(), polt(), gen()).value

    if result is not None:
        return MortalityTables(result).MortalityTable
    else:
        raise ValueError('MortTable not found')


def SurrRate(y):
    """Surrender Rate"""
    table = AsmpLookup.match("Surrender", prod(), polt(), gen()).value

    if table is None:
        raise ValueError('Surrender not found')

    result =  AssumptionTables.get((table, y), None)

    if result is None:
        return SurrRate(y-1)
    else:
        return result


# ---------------------------------------------------------------------------
# References

AsmpLookup = ("Interface", ("...", "Input", "AsmpLookup"), "auto")

AssumptionTables = ("Pickle", 2325077011464)

MortalityTables = ("Pickle", 2325070426248)

gen = ("Interface", ("..", "Policy", "Gen"), "auto")

polt = ("Interface", ("..", "Policy", "PolicyType"), "auto")

prod = ("Interface", ("..", "Policy", "Product"), "auto")

sex = ("Interface", ("..", "Policy", "Sex"), "auto")


from modelx.serialize.jsonvalues import *

_name = "fastlife"

_allow_none = False

_spaces = [
    "Input",
    "LifeTable",
    "Economic",
    "PV",
    "Projection"
]

# ---------------------------------------------------------------------------
# References

pd = ("Module", "pandas")

np = ("Module", "numpy")


"""Source module to create ``Assumption`` space from.

This module is a source module to create ``Assumption`` space and its
sub spaces from.
The formulas of the cells in the ``Assumption`` space are created from the
functions defined in this module.

The ``Assumption`` space is the base space of the assumption spaces
for individual policies, which are derived from and belong to
the ``Assumption`` space as its dynamic child spaces.

The assumption spaces for individual policies are parametrized by ``PolicyID``.
For example, to get the assumption space of the policy whose ID is 171::

    >> asmp = model.Assumption(171)

The cells in an assumption space for each individual policy retrieve
input data, calculate and hold values of assumptions specific to that policy,
so various spaces in :mod:`Input<simplelife.build_input>` must be accessible
from the ``Assumption`` space.

.. rubric:: Project Templates

This module is included in the following project templates.

* :mod:`simplelife`
* :mod:`nestedlife`

.. rubric:: Referred Spaces

The ``Assumption`` space and its sub spaces depend of the following spaces.
See references sections below for aliases to those spaces and their members
that are referenced in the ``Assumption`` spaces.

* :mod:`Policy<simplelife.policy>` its sub spaces
* ``LifeTable`` in :mod:`Input<simplelife.build_input>`
* ``MortalityTables`` in :mod:`Input<simplelife.build_input>`
* ``Assumption`` in :mod:`Input<simplelife.build_input>`

.. rubric:: Space Parameters

Attributes:
    PolicyID: Policy ID

.. rubric:: References in Base

Attributes:
    asmp_tbl: ``AssumptionTables`` space in :mod:`Input<simplelife.build_input>` space
    asmp: ``Assumption`` space in :mod:`Input<simplelife.build_input>` space
    MortalityTables: ``MortalityTables`` space in :mod:`Input<simplelife.build_input>` space

.. rubric:: References in Sub

Attributes:
    pol: Alias to :mod:`Policy[PolicyID]<simplelife.policy>`
    prod: Alias to :attr:`Policy[PolicyID].Product<simplelife.policy.Product>`
    polt: Alias to :attr:`Policy[PolicyID].PolicyType<simplelife.policy.PolicyType>`
    gen: Alias to :attr:`Policy[PolicyID].Gen<simplelife.policy.Gen>`



"""

policy_attrs = []

# #--- Mortality ---

def MortTable():
    """Mortality Table"""
    result = asmp.BaseMort.match(prod, polt, gen).value

    if result is not None:
        return MortalityTables(result).MortalityTable
    else:
        raise ValueError('MortTable not found')


def LastAge():
    """Age at which mortality becomes 1"""
    x = 0
    while True:
        if BaseMortRate(x) == 1:
            return x
        x += 1
            

def BaseMortRate(x):
    """Bae mortality rate"""
    return MortTable()(pol.Sex, x)


def MortFactor(y):
    """Mortality factor"""
    table = asmp.MortFactor.match(prod, polt, gen).value

    if table is None:
        raise ValueError('MortFactor not found')

    result = asmp_tbl.cells[table](y)

    if result is None:
        return MortFactor(y-1)
    else:
        return result

# --- Surrender Rates ---
def SurrRate(y):
    """Surrender Rate"""
    table = asmp.Surrender.match(prod, polt, gen).value

    if table is None:
        raise ValueError('Surrender not found')

    result =  asmp_tbl.cells[table](y)

    if result is None:
        return SurrRate(y-1)
    else:
        return result

# --- Commissions ---
def CommInitPrem():
    """Initial commission per premium"""
    result = asmp.CommInitPrem.match(prod, polt, gen).value

    if result is not None:
        return result
    else:
        raise ValueError('CommInitPrem not found')


def CommRenPrem():
    """Renewal commission per premium"""
    result = asmp.CommRenPrem.match(prod, polt, gen).value

    if result is not None:
        return  result
    else:
        raise ValueError('CommRenPrem not found')

def CommRenTerm():
    """Renewal commission term"""
    result = asmp.CommRenTerm.match(prod, polt, gen).value

    if result is not None:
        return result
    else:
        raise ValueError('CommRenTerm not found')

# # --- Expenses ---
def ExpsAcqSA():
    """Acquisition expense per sum assured"""
    return asmp.ExpsAcqSA.match(prod, polt, gen).value

def ExpsAcqAnnPrem():
    """Acquisition expense per annualized premium"""
    return asmp.ExpsAcqAnnPrem.match(prod, polt, gen).value

def ExpsAcqPol():
    """Acquisition expense per policy"""
    return asmp.ExpsAcqPol.match(prod, polt, gen).value

def ExpsMaintSA():
    """Maintenance expense per sum assured"""
    return asmp.ExpsMaintSA.match(prod, polt, gen).value

def ExpsMaintAnnPrem():
    """Maintenance expense per annualized premium"""
    return asmp.ExpsMaintPrem.match(prod, polt, gen).value

def ExpsMaintPol():
    """Maintenance expense per policy"""
    return asmp.ExpsMaintPol.match(prod, polt, gen).value

def CnsmpTax():
    """Consumption tax rate"""
    return asmp.CnsmpTax()

def InflRate():
    """Inflation rate"""
    return asmp.InflRate()




    
    















"""
:mod:`solvency2` Life SCR radar chart
=====================================

This script draws radar charts for selected policies.
Each spoke of the radar chart represents a sub risk of SCR life risks,
and the sizes of the sub risks by duration are drawn in the radar chart. 

"""
import modelx as mx
import pandas as pd
from draw_charts_radar import draw_radar

model = mx.read_model("model")
scr = model.SCR_life
risks = ('mort', 'longev', 'disab', 'exps', 'lapse')
scenid = 1


def draw(polid):

    data = {}
    for t in range(0, 20, 5):
        data['t=' + str(t)] = scr[t, polid, scenid].Life.to_series(risks)
    
    draw_radar(pd.DataFrame(data), 
               ax_title='Policy ID: ' + str(polid),
               fig_title='SCR Life Risks')

for i in (41, 171):
    draw(i)





"""Economic scenarios

The ``Economic`` spaces provides economic assumptions such as
interest rate scenarios.

This Space is included in:

* :mod:`simplelife`
* :mod:`nestedlife`
* :mod:`ifrs17sim`
* :mod:`solvency2`


.. _ExcelRange:
   https://docs.modelx.io/en/latest/reference/dataclient.html#excelrange

.. rubric:: Parameters

``Economic`` Space is parameterized with :attr:`ScenID`::

        >>> simplelife.Economic.parameters
        ('ScenID',)

Each ItemSpace represents economic scenarios for a specific :attr:`ScenID`.
For example, ``Economic[1]`` contains economic scenarios for ScenID 1.

Attributes:
    ScenID(:obj:`int`): Scenario ID

.. rubric:: References

Attributes:
    Scenario: `ExcelRange`_ object holding the data of interest rate
        assumptions. The data is read from *Scenarios* range in *input.xlsx*.

Example:

    An example of ``Economic`` in the :mod:`simplelife` model::

        >>> simplelife.Economic[1].DiscRate(0)
        0.015

"""

from modelx.serialize.jsonvalues import *

_formula = lambda ScenID: None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def DiscRate(t):
    return Scenarios[ScenID, "IntRate", t]


def InflFactor(t):
    if t == 0:
        return 1
    else:
        return InflFactor(t-1) / (1 + AsmpLookup("InflRate"))


def InvstRetRate(t):
    return DiscRate(t)


# ---------------------------------------------------------------------------
# References

ScenID = 1

AsmpLookup = ("Interface", ("..", "Input", "AsmpLookup"), "auto")

Scenarios = ("Pickle", 2325094462088)


import sys
import os.path
import argparse
import shutil
from lifelib._dirs import (TEMPLATES, DEFAULT_TEMPLATE)


def get_argparser():
    parse = argparse.ArgumentParser(description="Create a new project.")
    parse.add_argument('proj_dir',
                       help="Path to the project folder")
    parse.add_argument('--template', default=DEFAULT_TEMPLATE,
                       help="Name of a library")
    return parse


def main(argv=sys.argv[1:]):
    """Create a new project."""

    args = vars(get_argparser().parse_args(argv))

    if args['template'] not in TEMPLATES:
        raise ValueError("Template %s not found" % args['template'])
    else:
        template = args['template']

    create(template, args['proj_dir'])

    return 0


def create(template=DEFAULT_TEMPLATE, path=None):
    """Create a copy of a library

    Args:
        template(:obj:`str`, optional): name of the library.
            Defaults to ``basiclife``.
        path(:obj:`str`, optional): path to the copied folder. Absolute path
            or relative path to the current working directory/folder.
            If omitted, ``template`` is used.
    """

    if not path:
        path = template

    proj_dir = os.path.abspath(path)
    tpath = TEMPLATES[template]
    shutil.copytree(tpath, proj_dir)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



import modelx as mx
import lifelib.projects.simplelife.model



import pathlib

model_path = pathlib.Path(lifelib.projects.simplelife.model.__file__).parent


model = mx.read_model(model_path)


def run_newlife():
    proj = model.Projection
    for i in range(10, 301, 10):
        print(i, proj(i).PV_NetCashflow(0))


if __name__ == '__main__':
    import timeit
    print(timeit.timeit('run_newlife()', number=1, globals=globals()))



"""Space for cashflow projection.

This Space is for projecting cashflows of individual model points.

.. rubric:: Inheritance Structure

The ``Projection`` Space inherits its contents from its
base Spaces,
:mod:`~simplelife.model.BaseProj` and :mod:`~simplelife.model.PV`.
Projection items are inherited from :mod:`~simplelife.model.BaseProj`
and the present values of the cashflow items are
inherited from :mod:`~simplelife.model.BaseProj`.

.. blockdiag::

   blockdiag {
     default_node_color="#D5E8D4";
     default_linecolor="#628E47";
     BaseProj[style=dotted]
     BaseProj <- OuterProj [hstyle=generalization]
     PresentValue[style=dotted]
     PresentValue <- OuterProj [hstyle=generalization];
   }

.. rubric:: Parameters

This Space is parametrized with ``PolicyID`` and ``ScenID``::

    >>> simplelife.Projection.parameters
    ('PolicyID', 'ScenID')

Calling this space with a pair of integers returns the ItemSpace
for the policy ID and scenario ID. ``ScenID`` has a default value of 1,
so for example ``Projection[1]`` represents the Projection Space for Policy 1.

Attributes:
    PolicyID(:obj:`int`): Policy ID
    ScenID(:obj:`int`, optional): Scenario ID, defaults to 1.

.. rubric:: Composition Structure

This Space has child Spaces,
:mod:`~simplelife.model.Projection.Policy` and :mod:`~simplelife.model.Projection.Assumptions`.
The :mod:`~simplelife.model.Projection.Policy` Space contains Cells representing policy attributes, such as
product type, issue age, sum assured, etc.
It also contains Cells for calculating policy values such as premium rates and
cash surrender value rates.
The :mod:`~simplelife.model.Projection.Assumptions` Space contains Cells to pick up assumption data for
its model point.

.. blockdiag::

   blockdiag {
     default_node_color="#D5E8D4";
     default_linecolor="#628E47";
     node_width=150;
     Proj[label="Projection\\n[PolicyID, ScenID=1]", stacked];
     Proj <- Assumptions [hstyle=composition];
     Proj <- Policy [hstyle=composition];
   }



.. rubric:: References

The following attributes are referenced in this Space by its base Spaces.

Attributes:
    pol: Alias for :mod:`~simplelife.model.Projection.Policy` child Space
    asmp: Alias for :mod:`~simplelife.model.Projection.Assumptions` child Space
    scen: Alias for :mod:`~simplelife.model.Economic` Space


"""

from modelx.serialize.jsonvalues import *

_formula = lambda PolicyID, ScenID=1: None

_bases = [
    ".BaseProj",
    ".PV"
]

_allow_none = None

_spaces = [
    "Policy",
    "Assumptions"
]

# ---------------------------------------------------------------------------
# Cells

def DiscRate(t):
    """Rates for discount cashflows

    Refers to :func:`Economic[ScenID].DiscRate<simplelife.model.Economic.DiscRate>`
    """
    return scen[ScenID].DiscRate(t)


def InflFactor(t):
    """Inflation factors to adjust expense cashflows

    Refers to :func:`Economic[ScenID].InflFactor<simplelife.model.Economic.InflFactor>`
    """
    return scen[ScenID].InflFactor(t)


def InvstRetRate(t):
    """Rate of investment return

    Refers to :func:`Economic[ScenID].InvstRetRate<simplelife.model.Economic.InvstRetRate>`
    """
    return scen[ScenID].InvstRetRate(t)


# ---------------------------------------------------------------------------
# References

pol = ("Interface", (".", "Policy"), "auto")

asmp = ("Interface", (".", "Assumptions"), "auto")

scen = ("Interface", ("..", "Economic"), "auto")


"""
Draw radar charts
=================

This module contains :func:`draw_radar`, a function to draw radar charts a.k.a
spyder charts.

The code in this module is derived from the matplot example,
`Radar chart (aka spider or star chart) <https://matplotlib.org/gallery/specialty_plots/radar_chart.html>`_.

See `Matplotlib license <https://matplotlib.org/users/license.html>`_.
"""

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
from matplotlib.path import Path
from matplotlib.spines import Spine
from matplotlib.projections.polar import PolarAxes
from matplotlib.projections import register_projection


def radar_factory(num_vars, frame='circle'):
    """Create a radar chart with `num_vars` axes.

    This function creates a RadarAxes projection and registers it.

    Parameters
    ----------
    num_vars : int
        Number of variables for radar chart.
    frame : {'circle' | 'polygon'}
        Shape of frame surrounding axes.

    """
    # calculate evenly-spaced axis angles
    theta = np.linspace(0, 2*np.pi, num_vars, endpoint=False)

    def draw_poly_patch(self):
        # rotate theta such that the first axis is at the top
        verts = unit_poly_verts(theta + np.pi / 2)
        return plt.Polygon(verts, closed=True, edgecolor='k')

    def draw_circle_patch(self):
        # unit circle centered on (0.5, 0.5)
        return plt.Circle((0.5, 0.5), 0.5)

    patch_dict = {'polygon': draw_poly_patch, 'circle': draw_circle_patch}
    if frame not in patch_dict:
        raise ValueError('unknown value for `frame`: %s' % frame)

    class RadarAxes(PolarAxes):

        name = 'radar'
        # use 1 line segment to connect specified points
        RESOLUTION = 1
        # define draw_frame method
        draw_patch = patch_dict[frame]

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # rotate plot such that the first axis is at the top
            self.set_theta_zero_location('N')

        def fill(self, *args, closed=True, **kwargs):
            """Override fill so that line is closed by default"""
            return super().fill(closed=closed, *args, **kwargs)

        def plot(self, *args, **kwargs):
            """Override plot so that line is closed by default"""
            lines = super().plot(*args, **kwargs)
            for line in lines:
                self._close_line(line)

        def _close_line(self, line):
            x, y = line.get_data()
            # FIXME: markers at x[0], y[0] get doubled-up
            if x[0] != x[-1]:
                x = np.concatenate((x, [x[0]]))
                y = np.concatenate((y, [y[0]]))
                line.set_data(x, y)

        def set_varlabels(self, labels):
            self.set_thetagrids(np.degrees(theta), labels)

        def _gen_axes_patch(self):
            return self.draw_patch()

        def _gen_axes_spines(self):
            if frame == 'circle':
                return super()._gen_axes_spines()
            # The following is a hack to get the spines (i.e. the axes frame)
            # to draw correctly for a polygon frame.

            # spine_type must be 'left', 'right', 'top', 'bottom', or `circle`.
            spine_type = 'circle'
            verts = unit_poly_verts(theta + np.pi / 2)
            # close off polygon by repeating first vertex
            verts.append(verts[0])
            path = Path(verts)

            spine = Spine(self, spine_type, path)
            spine.set_transform(self.transAxes)
            return {'polar': spine}

    register_projection(RadarAxes)
    return theta


def unit_poly_verts(theta):
    """Return vertices of polygon for subplot axes.

    This polygon is circumscribed by a unit circle centered at (0.5, 0.5)
    """
    x0, y0, r = [0.5] * 3
    verts = [(r*np.cos(t) + x0, r*np.sin(t) + y0) for t in theta]
    return verts


def get_rgrids(max_value, rgrid_count):
    """Calculate radial grid steps and return it as a list"""

    import math

    digits = math.floor(math.log(max_value, 10))
    scale = max_value / (10**digits)

    ubound = math.ceil(scale) * (10**digits)
    step = ubound / (rgrid_count+1)

    return list(step * i for i in range(1, rgrid_count+1))


def draw_radar(df: pd.DataFrame, ax_title=None, fig_title=None,
               colors=('b', 'r', 'g', 'm', 'y')):
    """Draw radar chart"""

    spoke_count, line_count = df.shape

    theta = radar_factory(spoke_count, frame='polygon')
    spoke_labels = list(df.index)

    fig, ax = plt.subplots(subplot_kw=dict(projection='radar'))
    fig.subplots_adjust(wspace=0.25, hspace=0.20, top=0.85, bottom=0.05)

    if len(colors) < line_count:
        colors = (colors * (line_count // len(colors) + 1))[:line_count]

    ax.set_rgrids(get_rgrids(df.max().max(), 4))
    if ax_title:
        ax.set_title(ax_title, weight='normal', size='medium',
                     position=(0.5, 1.09),
                     horizontalalignment='center', verticalalignment='center')

    for col, color in zip(df, colors):
        ax.plot(theta, list(df[col]), color=color)
        ax.fill(theta, list(df[col]), facecolor=color, alpha=0.25)

    ax.set_varlabels(spoke_labels)

    # add legend relative to top-left plot
    labels = list(df.columns)
    legend = ax.legend(labels, loc=(0.9, .95),
                       labelspacing=0.1)

    fig.text(0.5, 0.965, fig_title,
             horizontalalignment='center', color='black', weight='bold',
             size='large')

    return ax


if __name__ == '__main__':

    data = {'Line 1': [0.88, 1.30, 0.50, 0.03],
            'Line 2': [0.07, 0.95, 0.04, 0.05],
            'Line 3': [0.01, 0.02, 0.85, 0.19],
            'Line 4': [0.02, 0.01, 0.07, 0.60],
            'Line 5': [0.01, 0.01, 0.02, 0.71]}

    index = ['Sulfate', 'Nitrate', 'EC', 'OC1']

    df = pd.DataFrame(data=data, index=index)

    draw_radar(df, fig_title='Sample Radar')

    plt.show()



"""Policy attributes and policy values

This Space is a child Space of :mod:`~nestedlife.model.OuterProj`.
A child Space derived from this Space is created under each
:mod:`~nestedlife.model.OuterProj` ItemSpace,
such as ``nestedlife.OuterProj[1].Policy``,
which holds policy attributes and policy values for Policy 1.
The Space is used by :mod:`~nestedlife.model.BaseProj.Assumptions`
and :mod:`~nestedlife.model.OuterProj.InnerProj`.

Some Cells in this Space, such as :func:`Product` and :func:`IssueAge`,
are for retrieving attributes for the selected policy from :attr:`PolicyData`.
Some other Cells, such as :func:`GrossPremRate` are for
calculating policy values for the policy from the attributes and
product specs looked up through :attr:`SpecLookup`.


.. blockdiag::

   blockdiag {
     default_node_color="#D5E8D4";
     default_linecolor="#628E47";
     node_width=150;
     Proj[label="OuterProj\\n[PolicyID, ScenID=1]", stacked];
     Proj <- Assumptions [hstyle=composition];
     Proj <- Policy [hstyle=composition];
   }


.. rubric:: Parameters

Since :mod:`~nestedlife.model.OuterProj` is parameterized with
:attr:`PolicyID` and :attr:`ScenID`, this Space is also
parameterized as a child space of :mod:`~nestedlife.model.OuterProj`.
For example, ``nestedlife.Projection[1].Policy.GrossPremRate()``
represents the gross premium rate for Policy 1.

Attributes:
    PolicyID(:obj:`int`): Policy ID
    ScenID(:obj:`int`, optional): Scenario ID, defaults to 1.


.. rubric:: References

Attributes:
    LifeTable: :mod:`~nestedlife.model.LifeTable` Space
    PolicyData: `ExcelRange`_ object holding data read from the
        Excel range *PolicyData* in *input.xlsx*.
    SpecLookup: :func:`~nestedlife.model.Input.SpecLookup`
    PremTerm: Alias for :func:`PolicyTerm`

.. _ExcelRange:
   https://docs.modelx.io/en/latest/reference/dataclient.html#excelrange

"""

from modelx.serialize.jsonvalues import *

_formula = None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def AnnPremRate():
    """Annualized Premium Rate per Sum Assured"""
    return GrossPremRate * (1/10 if PremFreq == 0 else PremFreq)


def CashValueRate(t):
    """Cash Value Rate per Sum Assured"""
    return max(ReserveNLP_Rate('PREM', t) - SurrCharge(t), 0)


def GrossPremRate():
    """Gross Premium Rate per Sum Assured per payment"""

    alpha = LoadAcqSA()
    beta = LoadMaintPrem()
    gamma = LoadMaintSA()
    gamma2 = LoadMaintSA2()
    delta = LoadMaintPremWaiverPrem()

    x, n, m = IssueAge(), PolicyTerm(), PremTerm()

    comf = LifeTable[Sex(), IntRate('PREM'), TableID('PREM')]

    if Product == 'TERM' or Product == 'WL':
        return (comf.Axn(x, n) + alpha + gamma * comf.AnnDuenx(x, n, PremFreq)
                + gamma2 * comf.AnnDuenx(x, n-m, 1, m)) / (1-beta-delta) / PremFreq / comf.AnnDuenx(x, m, PremFreq)

    elif Product == 'ENDW':
        return (comf.Exn(x, n) + comf.Axn(x, n) + alpha + gamma * comf.AnnDuenx(x, n, PremFreq)
                + gamma2 * comf.AnnDuenx(x, n-m, 1, m)) / (1-beta-delta) / PremFreq / comf.AnnDuenx(x, m, PremFreq)
    else:
        raise ValueError('invalid product')


def GrossPremTable():
    """Gross premium table"""
    return None


def InitSurrCharge():
    """Initial Surrender Charge Rate"""

    param1 = SpecLookup.match("SurrChargeParam1", Product(), PolicyType(), Gen()).value
    param2 = SpecLookup.match("SurrChargeParam2", Product(), PolicyType(), Gen()).value

    if param1 is None or param2 is None:
        raise ValueError('SurrChargeParam not found')

    return param1 + param2 * min(PolicyTerm / 10, 1)


def IntRate(RateBasis):
    """Interest Rate"""

    if RateBasis == 'PREM':
        basis = 'IntRatePrem'
    elif RateBasis == 'VAL':
        basis = 'IntRateVal'
    else:
        raise ValueError('invalid RateBasis')

    result = SpecLookup.match(basis, Product(), PolicyType(), Gen()).value

    if result is not None:
        return result
    else:
        raise ValueError('invalid RateBais')


def LoadAcqSA():
    """Acquisition Loading per Sum Assured"""
    param1 = SpecLookup("LoadAcqSAParam1", Product())
    param2 = SpecLookup("LoadAcqSAParam2", Product())

    return param1 + param2 * min(PolicyTerm / 10, 1)


def LoadMaintPrem():
    """Maintenance Loading per Gross Premium"""

    if SpecLookup("LoadMaintPremParam1", Product()) is not None:
        return SpecLookup("LoadMaintPremParam1", Product())

    elif SpecLookup("LoadMaintPremParam2", Product()) is not None:
        param = SpecLookup("LoadMaintPremParam2", Product())
        return (param + min(10, PolicyTerm)) / 100

    else:
        raise ValueError('LoadMaintPrem parameters not found')


def LoadMaintPremWaiverPrem():
    """Maintenance Loading per Gross Premium for Premium Waiver"""

    if PremTerm < 5:
        return 0.0005
    elif PremTerm < 10:
        return 0.001
    else:
        return 0.002


def LoadMaintSA():
    """Maintenance Loading per Sum Assured during Premium Payment"""

    result = SpecLookup.match("LoadMaintSA", Product(), PolicyType(), Gen()).value

    if result is not None:
        return result
    else:
        raise ValueError('lookup failed')


def LoadMaintSA2():
    """Maintenance Loading per Sum Assured after Premium Payment"""

    result = SpecLookup.match("LoadMaintSA2", Product(), PolicyType(), Gen()).value

    if result is not None:
        return result
    else:
        raise ValueError('lookup failed')


def NetPremRate(basis):
    """Net Premium Rate"""

    gamma2 = LoadMaintSA2
    comf = LifeTable[Sex(), IntRate(basis), TableID(basis)]

    x, n, m = IssueAge(), PolicyTerm(), PremTerm()

    if Product == 'TERM' or Product == 'WL':
        return (comf.Axn(x, n) + gamma2 * comf.AnnDuenx(x, n-m, 1, m)) / comf.AnnDuenx(x, n)

    elif Product == 'ENDW':
        return (comf.Axn(x, n) + gamma2 * comf.AnnDuenx(x, n-m, 1, m)) / comf.AnnDuenx(x, n)

    else:
        raise ValueError('invalid product')


def ReserveNLP_Rate(basis, t):
    """Net level premium reserve rate"""

    gamma2 = LoadMaintSA2()

    lt = LifeTable[Sex(), IntRate(basis), TableID(basis)]

    x, n, m = IssueAge(), PolicyTerm(), PremTerm()

    if t <= m:
        return lt.Axn(x+t, n-t) + gamma2 * lt.AnnDuenx(x+t, n-m, 1, m-t) \
                - NetPremRate(basis) * lt.AnnDuenx(x+t, m-t)
    else:
        return lt.Axn(x+t, n-t) + gamma2 * lt.AnnDuenx(x+t, n-m, 1, m-t)


def ReserveRate():
    """Valuation Reserve Rate per Sum Assured"""
    return None


def SurrCharge(t):
    """Surrender Charge Rate per Sum Assured"""
    m = PremTerm()
    return InitSurrCharge * max((min(m, 10) - t) / min(m, 10), 0)


def TableID(RateBasis):
    """Mortality Table ID"""

    if RateBasis == 'PREM':
        basis = "MortTablePrem"
    elif RateBasis == 'VAL':
        basis = "MortTableVal"
    else:
        raise ValueError('invalid RateBasis')

    result = SpecLookup.match(basis, Product(), PolicyType(), Gen()).value

    if result is not None:
        return result
    else:
        raise ValueError('invalid RateBais')


def UernPremRate():
    """Unearned Premium Rate"""
    return None


Product = lambda: PolicyData[PolicyID, 'Product']

PolicyType = lambda: PolicyData[PolicyID, 'PolicyType']

Gen = lambda: PolicyData[PolicyID, 'Gen']

Channel = lambda: PolicyData[PolicyID, 'Channel']

Sex = lambda: PolicyData[PolicyID, 'Sex']

Duration = lambda: PolicyData[PolicyID, 'Duration']

IssueAge = lambda: PolicyData[PolicyID, 'IssueAge']

PremFreq = lambda: PolicyData[PolicyID, 'PremFreq']

PolicyTerm = lambda: PolicyData[PolicyID, 'PolicyTerm']

PolicyCount = lambda: PolicyData[PolicyID, 'PolicyCount']

SumAssured = lambda: PolicyData[PolicyID, 'SumAssured']

# ---------------------------------------------------------------------------
# References

LifeTable = ("Interface", ("...", "LifeTable"), "auto")

PolicyData = ("Pickle", 2324875789768)

PremTerm = ("Interface", (".", "PolicyTerm"), "auto")

SpecLookup = ("Interface", ("...", "Input", "SpecLookup"), "auto")


"""The main space in the Smith-Wilson model

``SmithWilson`` is the only space in the Smith-Wilson model.

The Smith-Wilson method is used for extrapolating risk-free interest rates under the Solvency II framework.
The method is described in
*"QIS 5 Risk-free interest rates – Extrapolation method"*,
a technical paper issued by CEIOPS (the predecessor of EIOPA).
The technical paper is available on `EIOPA's web site`_.
Cells in this space are named consistently
with the mathematical symbols in `the technical paper`_.

.. rubric:: References

Attributes:
    log: `log <https://docs.python.org/3/library/math.html#math.log>`_ function from `the standard math library`_
    exp: `exp <https://docs.python.org/3/library/math.html#math.exp>`_ function from `the standard math library`_
    np: `Numpy`_ module
    N: Number of durations for which the observed spot rates are available
    spot_rates: List of the observed spot rates (annual compound)
    UFR: The ultimate forward rate (continuous compound)
    alpha: The convergence parameter :math:`\alpha`

.. _Numpy : https://numpy.org/

.. _EIOPA's web site: https://wayback.archive-it.org/org-1495/20191229100044/https:/eiopa.europa.eu/publications/qis/insurance/insurance-quantitative-impact-study-5/background-documents

.. _the technical paper: https://wayback.archive-it.org/org-1495/20191229100044/https:/eiopa.europa.eu/Publications/QIS/ceiops-paper-extrapolation-risk-free-rates_en-20100802.pdf

.. _the standard math library: https://docs.python.org/3/library/math.html
"""

from modelx.serialize.jsonvalues import *

_formula = None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def u(i):
    """Time (:math:`u_i`)

    :func:`u` is a series of discrete time points.
    ``i`` is a 1-based index. By default, :func:`u` just returns ``i``.

    For i = 1, ..., :attr:`N`, :func:`u` corresponds to
    :math:`u_i` in `the technical paper`_,
    The time to maturities of the observed zero-coupon bond prices.

    :attr:`spot_rates`, the observed spot rates must exist at each
    ``u(i)`` from i = 1 to :attr:`N`. Note that :attr:`spot_rates` is
    0-based, so the sport rate at ``u(i)`` is ``spot_rates[i-1]``.

    Args:
        i: The time index (1, 2, ...)

    """
    return i


def m(i):
    """Observed zero-coupon bond prices at time :math:`u_i`.

    :func:`m` is calculated from :attr:`spot_rates` as
    :math:`(1 + spot\_rates[i-1])^{-u_i}`

    Args:
        i(int): Time index (1, 2, ..., :attr:`N`)
    """
    return (1 + spot_rates[i-1]) ** (-u[i])


def mu(i):
    """Ultimate Forward Rate (UFR) discount factors

    :func:`mu` is defined as :math:`e^{-UFR\cdot u_i}`.

    Args:
        i(int): Time index (1, 2, ...)
    """
    return exp(-UFR * u[i])


def W(i, j):
    """The Wilson functions.

    :func:`W` corresponds to formula (2) on page 16 in `the technical paper`_
    defined as:

    .. math::

        W(t, u_j)=  \\
            e^{-UFR\cdot (t+u_j)}\cdot  \\
            \left\{   \\
                \\alpha\cdot\min(t, u_j) \\
                -0.5\cdot e^{-\\alpha\cdot\max(t, u_j)}\cdot(    \\
                    e^{\\alpha\cdot\min(t, u_j)}     \\
                    -e^{-\\alpha\cdot\min(t, u_j)}    \\
                    )   \\
            \\right\}

    where :math:`t = u_i`.

    Args:
        i(int): Time index (1, 2, ..., :attr:`N`)
        j(int): Time index (1, 2, ..., :attr:`N`)
    """
    t = u[i]
    uj = u[j]

    return exp(-UFR * (t+uj)) * (
            alpha * min(t, uj) - 0.5 * exp(-alpha * max(t, uj)) * (
                    exp(alpha*min(t, uj)) - exp(-alpha*min(t, uj))
            ))


def m_vector():
    """The :func:`m` vector.

    :func:`m` as 1-dimensional :attr:`N` length numpy array.
    """
    return np.array([m(i) for i in range(1, N+1)])


def mu_vector():
    """The :func:`mu` vector.

    :func:`mu` as 1-dimensional :attr:`N` length numpy array.
    """
    return np.array([mu(i) for i in range(1, N+1)])


def W_matrix():
    """The :func:`W` matrix.

    :func:`W` as a 2-dimensional :attr:`N` x :attr:`N` numpy array.
    """
    return np.array([[W(i, j) for j in range(1, N+1)] for i in range(1, N+1)])


def zeta_vector():
    """The :func:`zeta` vector.

    :func:`zeta_vector` returns :math:`\zeta` parameters calculated
    by formula (5) on page 17 in `the technical paper`_, which is

    .. math::

        \\bf \zeta= W^{-1}(p-\mu)

    """
    return np.linalg.inv(W_matrix()) @ (m_vector() - mu_vector())


def zeta(i):
    """The :math:`\zeta_i` parameters fitted to the observed spot rates.

    Args:
        i(int): Time index (1, 2, ..., :attr:`N`)
    """
    return zeta_vector()[i-1]


def P(i):
    """Zero-coupon bond prices extrapolated by the Smith-Wilson method.

    :func:`P` corresponds to formula (1) on page 16 or formula (6)
    on page 18 in `the technical paper`_, defined as:

    .. math::

        P(t) = e^{-UFR\cdot t}+\sum_{j=1}^{N}\zeta_{j}\cdot W(t, u_j)

    substituting :math:`t` with :math:`u_i`.

    The values of :func:`P` for ``i=1, ..., N`` should be the same
    as the values of :func:`m`, the observed bond prices.

    Args:
        i(int): Time index (1, 2, ...)
    """
    return mu(i) + sum(zeta(j) * W(i, j) for j in range(1, N+1))


def R(i):
    """The extrapolated annual compound sport rates.

    :func:`R` corresponds to :math:`R_t` defined as:

    .. math::

        R_t = \left(\\frac{1}{P(t)}\\right)^\left(\\frac{1}{t}\\right)-1

    on page 18 in `the technical paper`_,
    substituting :math:`t` with :math:`u_i`.

    The values of :func:`R` for ``i=1,...,N`` should be same as the values
    of the observed spot rates :attr:`spot_rates` for 0, ... ,N-1.

    Args:
        i(int): Time index (1, 2, ...)
    """
    return (1 / P(i)) ** (1 / u(i)) - 1


# ---------------------------------------------------------------------------
# References

np = ("Module", "numpy")

log = ("Pickle", 2074070178376)

exp = ("Pickle", 2074070177016)

UFR = 0.028587456851912472

alpha = 0.128562

N = 25

spot_rates = ("Pickle", 2072439860616)


import sys, os, pickle
from lifelib.projects.simplelife.scripts import simplelife
from lifelib.tests.data import round_signif

filepath = os.path.join(os.path.dirname(__file__), 'data_simplelife')

if '' not in sys.path:
    sys.path.insert(0, '')


def generate_data(model):
    data = []
    proj = model.Projection
    for i in range(10, 301, 10):
        data.append(round_signif(proj(i).PV_NetCashflow(0), 10))

    with open(filepath, 'wb') as file:
        pickle.dump(data, file, protocol=4)


if __name__ == '__main__':
    generate_data(simplelife.build())



"""Override module for the mortality/longevity risk calculation

The formulas in this module overrides cells related to mortality in
:mod:`projection <solvency2.projection>` module.
"""

from modelx.serialize.jsonvalues import *

_formula = None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def MortRateFactor(t):
    """Mortality rate factor applied from time ``t0`` and there after"""
    if t >= t0:        

        if AttAge(t) < asmp.LastAge():

            if Risk == 'mort':
                return 1 + Factor(Risk, Shock, Scope)
            elif Risk == 'longev':
                return 1 - Factor(Risk, Shock, Scope)
            else:
                ValueError("invalid Risk: %s" % Risk)

        else:
            return 1
    else:
        return 1


def PolsDeath(t):
    """Number of policies: Death override"""

    return (PolsIF_Beg1(t) * asmp.BaseMortRate(AttAge(t)) 
            * asmp.MortFactor(t) * MortRateFactor(t))





from modelx.serialize.jsonvalues import *

def _formula(Risk='base', Shock=None, Scope=None):


    if Risk == 'mort' or Risk == 'longev':
        bases = [_space.model.Override.Mortality, _space]  

    elif Risk == 'lapse':
        if Shock == 'mass':
            bases = [_space.model.Override.LapseMass, _space]
        else:
            bases = [_space.model.Override.Lapse, _space]
    elif Risk == 'exps':
            bases = [_space.model.Override.Expense, _space]   
    else:
        bases = [_space]

    refs = {'pol': Policy[PolicyID],
            'asmp': Assumptions[PolicyID],
            'scen': Economic[ScenID],
            'DiscRate': Economic[ScenID].DiscRate,
            'Factor': _space.model.Input.Factor}

    return {'bases':bases, 'refs': refs}


_bases = [
    "..BaseProj",
    "..PV"
]

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# References

Economic = ("Interface", ("...", "Economic"), "auto")

Policy = ("Interface", ("...", "Policy"), "auto")

Assumptions = ("Interface", ("...", "Assumptions"), "auto")


"""
Dynamic policy decrement
===================================

As explained in :doc:`/libraries/notebooks/savings/savings_example2`,
most variable annuity models include dynamic lapse assumptions.
A dynamic lapse assumption is used to reflect such policy holder's behaviour
that they are more likely to terminate their policies
when their account values are larger in comparison to the guaranteed amounts
of their policies than they are when the account values are smaller than the guaranteed amounts.

The graph shows deterministic lapse rates and policy decrement
based on the lapse rates. The lates and derement do not vary by scenarios.
"""
import numpy as np
import pandas as pd
import modelx as mx
import matplotlib.pyplot as plt


model = mx.read_model('CashValue_ME_EX2')
proj = model.Projection

pd.Series(proj[4].lapse_rate(t).loc[1][1] for t in range(120)).plot.line(
    legend=False, grid=True, title='Lapse rate by duration')

plt.subplots()
pd.Series(proj[4].pols_if_at(t, 'BEF_DECR').loc[1][1] for t in range(120)).plot.line(
    legend=False, grid=True, title='Number of policies')

#%%
# The graph below shows lapse rates and policy decrement for the first 100 scenarios,
# when a dynamic lapse is factored into the assumption.
# The lapse rates vary by in-the-moneyness of the account value and
# the number of remaining policies also vary by scenarios.

df = pd.DataFrame({t: proj[5].lapse_rate(t).loc[1][:100] for t in range(120)})
df.transpose().plot.line(legend=False, grid=True, title='Lapse rate by duration')

df = pd.DataFrame({t: proj[5].pols_if_at(t, 'BEF_DECR').loc[1][:100] for t in range(120)})
df.transpose().plot.line(legend=False, grid=True, title='Number of policies')

model.close()

#%%
# .. seealso::
#
#    * :doc:`/libraries/notebooks/savings/savings_example2` notebook in the :mod:`~savings` library
#


r"""
Account value paths
===================================

Account value paths generated by the Monte Carlo simulation.
The first 100 and 1000 paths out of 10000 are plotted.

The investment return on account value is modeled to follow the following geometric Brownian motion

.. math::

    \frac{\Delta{S}}{S}=r\Delta{t}+\sigma\epsilon\sqrt{\Delta{t}}

where :math:`\Delta{S}` is the change in the account value :math:`S`, in a small interval of time,
:math:`\Delta{t}`, :math:`r` is the constant risk free rate, :math:`\sigma`
is the volatility of the account value and :math:`\epsilon` is a random number
drawn from the standard normal distribution.

.. seealso::

    * :doc:`/libraries/notebooks/savings/savings_example1` notebook in the :mod:`~savings` library


"""
import modelx as mx
import pandas as pd

model = mx.read_model("CashValue_ME_EX1")
proj = model.Projection
av = proj.av_pp_at

avs = list(av(t, 'BEF_FEE') for t in range(proj.max_proj_len()))
df = pd.DataFrame(avs)[1]

for scen_size in [100, 1000]:
    df[range(1, scen_size + 1)].plot.line(
        legend=False, grid=True, title="%s scenarios" % scen_size)






"""Override module for the mass lapse risk calculation

The formulas in this module overrides cells related to lapse in
:mod:`projection <solvency2.projection>` module.
"""

def BenefitSurr(t):
    """Surrender benefits"""
    return SizeBenefitSurr(t) * (PolsSurr(t) + PolsSurrMass(t))


def PolsSurrMass(t):
    """Number of policies: Surrender"""
    factor = Factor(Risk, Shock, Scope) if t == t0 else 0

    return (PolsIF_Beg(t) + PolsRenewal(t) + PolsNewBiz(t)) * factor


def PolsIF_Beg1(t):
    """Number of policies: Beginning of period 1"""
    return PolsIF_Beg(t) + PolsRenewal(t) + PolsNewBiz(t) - PolsSurrMass(t)



"""Base Space for the :mod:`~ifrs17sim.model.OuterProj` and :mod:`~ifrs17sim.model.OuterProj.InnerProj`.

This Space serves as a base Space for :mod:`~ifrs17sim.model.OuterProj`
and :mod:`~ifrs17sim.model.OuterProj.InnerProj`,
and it contains Cells for cashflow projection.

.. rubric:: Inheritance Structure

.. blockdiag::

   blockdiag {
     default_node_color="#D5E8D4";
     default_linecolor="#628E47";
     BaseProj[style=dotted]
     BaseProj <- OuterProj [hstyle=generalization]
     PV[style=dotted]
     PV <- OuterProj [hstyle=generalization];
   }

.. blockdiag::

   blockdiag {
     default_node_color="#D5E8D4";
     default_linecolor="#628E47";
     BaseProj[style=dotted]
     BaseProj <- InnerProj [hstyle=generalization]
     PV[style=dotted]
     PV <- InnerProj [hstyle=generalization];
   }

``Pols``:
    Cells whose names start with ``Pols`` deal with number of policies.
    For example, ``PolsDeath(t)`` represents number of deaths between
    time ``t`` and ``t+1``.

``Size``:
    Cells whose names start with ``Size`` represents an amount per policy.
    For example, ``SizeBenefitDeath`` represents sum assured per policy.

``Exps``:
    Cells whose names start with ``Exps`` represents expense cashflows.
    For example, ``ExpsCommRen`` means the renewal commission cashflow.

``Benefit``:
    Cells whose names start with ``Benefit`` represents benefit cashflows.
    For example, ``BenefitDeath(t)`` death benefits incurred
    between ``t`` and ``t+1``.

"""

from modelx.serialize.jsonvalues import *

_formula = None

_bases = []

_allow_none = None

_spaces = [
    "Assumptions"
]

# ---------------------------------------------------------------------------
# Cells

def AccumCF(t):
    """Accumulated cashflows"""
    if t == 0:
        return 0
    else:
        return (AccumCF(t-1)
                + IntAccumCF(t-1)
                + NetInsurCF(t-1))


def AttAge(t):
    """Attained age at time ``t``"""
    return pol.IssueAge + t


def BenefitAccDth(t):
    """Accidental death benefits"""
    return SizeBenefitAccDth(t) * PolsAccDeath(t)


def BenefitAccHosp(t):
    """Accidental hospitalization benefits"""
    return SizeBenefitAccHosp(t) * PolsAccHosp(t)


def BenefitAnn(t):
    """Annuity benefits"""
    return SizeBenefitAnn(t) * PolsAnnuity(t)


def BenefitDeath(t):
    """Death benefits"""
    return SizeBenefitDeath(t) * PolsDeath(t)


def BenefitLiving(t):
    """Living benefits"""
    return SizeBenefitLiving(t) * PolsLiving(t)


def BenefitMat(t):
    """Matuirty benefits"""
    return SizeBenefitMat(t) * PolsMaturity(t)


def BenefitOther(t):
    """Other benefits"""
    return SizeBenefitOther(t) * PolsOther(t)


def BenefitSickHosp(t):
    """Sickness hospitalization benefits"""
    return SizeBenefitSickHosp(t) * PolsSickHosp(t)


def BenefitSurg(t):
    """Surgery benefits"""
    return SizeBenefitSurg(t) * PolsSurg(t)


def BenefitSurr(t):
    """Surrender benefits"""
    return SizeBenefitSurr(t) * PolsSurr(t)


def BenefitTotal(t):
    """Benefit Total"""
    return (BenefitMat(t)
            + BenefitDeath(t)
            + BenefitAccDth(t)
            + BenefitSurr(t)
            + BenefitAnn(t)
            + BenefitAccHosp(t)
            + BenefitSickHosp(t)
            + BenefitSurg(t)
            + BenefitLiving(t)
            + BenefitOther(t))


def ChangeRsrv(t):
    """Change in reserve"""
    return ReserveTotal_End(t+1) - ReserveTotal_End(t)


def ExpsAcq(t):
    """Acquisition expenses"""
    return SizeExpsAcq(t) * (PolsNewBiz(t) + PolsRenewal(t))


def ExpsAcqTotal(t):
    """Commissions and acquisition expenses"""
    return ExpsCommTotal(t) + ExpsAcq(t)


def ExpsCommInit(t):
    """Initial commissions"""
    return SizeExpsCommInit(t) * PolsIF_Beg1(t)


def ExpsCommRen(t):
    """Renewal commissions"""
    return SizeExpsCommRen(t) * PolsIF_Beg1(t)


def ExpsCommTotal(t):
    """Commissions Total"""
    return ExpsCommInit(t) + ExpsCommRen(t)


def ExpsMaint(t):
    """Maintenance expenses"""
    return SizeExpsMaint(t) * PolsIF_Beg1(t)


def ExpsMaintTotal(t):
    """Total maintenance expenses including other expenses"""
    return ExpsMaint(t) + ExpsOther(t)


def ExpsOther(t):
    """Other expenses"""
    return 0


def ExpsTotal(t):
    """Total expenses"""
    return (ExpsCommInit(t)
            + ExpsCommRen(t)
            + ExpsAcq(t)
            + ExpsMaint(t)
            + ExpsOther(t))


def IncomeTotal(t):
    """Income Total"""
    return PremIncome(t) + InvstIncome(t)


def InsurIF_Beg1(t):
    """Insurance in-force: Beginning of period 1"""
    return PolsIF_Beg1(t) * SizeSumAssured(t)


def InsurIF_End(t):
    """Insurance in-force: End of period"""
    return PolsIF_End(t) * SizeSumAssured(t)


def IntAccumCF(t):
    """Intrest on accumulated cashflows"""
    return (AccumCF(t)
            + PremIncome(t)
            - ExpsTotal(t)) * DiscRate(t)


def InvstIncome(t):
    """Investment income"""
    return SizeInvstIncome(t) * PolsIF_Beg1(t)


def NetInsurCF(t):
    """Net liability cashflow"""
    return (PremIncome(t)
            - BenefitTotal(t)
            - ExpsTotal(t))


def PolsAccDeath(t):
    """Number of policies: Accidental death"""
    return 0


def PolsAccHosp(t):
    """Number of policies: Accidental Hospitalization"""
    return 0


def PolsAnnuity(t):
    """Number of policies: Annuity"""
    return 0


def PolsDeath(t):
    """Number of policies: Death"""
    return PolsIF_Beg1(t) * asmp.BaseMortRate(AttAge(t)) * asmp.MortFactor(t)


def PolsIF_AftMat(t):
    """Number of policies: Maturity"""
    return PolsIF_End(t) - PolsMaturity(t)


def PolsIF_Beg(t):
    """Number of policies: Beginning of period"""
    return PolsIF_AftMat(t)


def PolsIF_Beg1(t):
    """Number of policies: Beginning of period 1"""
    return PolsIF_Beg(t) + PolsRenewal(t) + PolsNewBiz(t)


def PolsIF_End(t):
    """Number of policies: End of period"""
    if t == 0:
        return 0 # pol.PolicyCount
    else:
        return PolsIF_Beg1(t-1) - PolsDeath(t-1) - PolsSurr(t-1)


def PolsLiving(t):
    """Number of policies: Living benefits"""
    return 0


def PolsMaturity(t):
    """Number of policies: Maturity"""
    if t == pol.PolicyTerm:
        return PolsIF_End(t)
    else:
        return 0


def PolsNewBiz(t):
    """Number of policies: New business"""
    return pol.PolicyCount if t == 0 else 0


def PolsOther(t):
    """Number of policies: Other benefits"""
    return 0


def PolsRenewal(t):
    """Number of policies: Renewal policies"""
    return 0


def PolsSickHosp(t):
    """Number of policies: Sickness Hospitalization"""
    return 0


def PolsSurg(t):
    """Number of policies: Surgery"""
    return 0


def PolsSurr(t):
    """Number of policies: Surrender"""
    return PolsIF_Beg1(t) * asmp.SurrRate(t) * asmp.SurrRateMult(t)


def PremIncome(t):
    """Premium income"""
    return SizePremium(t) * PolsIF_Beg1(t)


def ProfitBefTax(t):
    """Profit before Tax"""

    return (PremIncome(t)
            + InvstIncome(t)
            - BenefitTotal(t)
            - ExpsTotal(t)
            - ChangeRsrv(t))


def ReserveHospRsrvEnd(t):
    """Hospitalization reserve: End of period"""
    return 0


def ReservePremRsrvEnd(t):
    """Premium reserve: End of period"""
    return SizeReservePremRsrvEnd(t) * PolsIF_End(t)


def ReserveTotal_End(t):
    """Total reserve: End of period"""
    return (ReservePremRsrvEnd(t)
            + ReserveUernPremEnd(t)
            + ReserveHospRsrvEnd(t))


def ReserveUernPremEnd(t):
    """Unearned Premium: End of period"""
    return 0


def SizeAnnPrem(t):
    """Annualized premium per policy at time ``t``"""
    return SizeSumAssured(t) * pol.AnnPremRate


def SizeBenefitAccDth(t):
    """Accidental death benefit per policy"""
    return 0


def SizeBenefitAccHosp(t):
    """Accidental hospitalization benefit per policy"""
    return 0


def SizeBenefitAnn(t):
    """Annuity benefit per policy"""
    return 0


def SizeBenefitDeath(t):
    """Death benefit per policy"""
    return SizeSumAssured(t)


def SizeBenefitLiving(t):
    """Living benefit per policy"""
    return 0


def SizeBenefitMat(t):
    """Maturity benefit per policy"""
    return 0


def SizeBenefitOther(t):
    """Other benefit per policy"""
    return 0


def SizeBenefitSickHosp(t):
    """Sickness hospitalization benefit per policy"""
    return 0


def SizeBenefitSurg(t):
    """Surgery benefit per policy"""
    return 0


def SizeBenefitSurr(t):
    """Surrender benefit per policy"""
    return SizeSumAssured(t) * (pol.CashValueRate(t)
                                + pol.CashValueRate(t+1)) / 2


def SizeExpsAcq(t):
    """Acquisition expense per policy at time t"""
    if t == 0:
        return (SizeAnnPrem(t) * asmp.ExpsAcqAnnPrem
                + (SizeSumAssured(t) * asmp.ExpsAcqSA + asmp.ExpsAcqPol)
                * InflFactor(t) / InflFactor(0))
    else:
        return 0


def SizeExpsCommInit(t):
    """Initial commission per policy at time t"""
    if t == 0:
        return SizePremium(t) * asmp.CommInitPrem * (1 + asmp.CnsmpTax)
    else:
        return 0


def SizeExpsCommRen(t):
    """Renewal commission per policy at time t"""
    if t == 0:
        return 0
    elif t < asmp.CommRenTerm:
        return SizePremium(t) * asmp.CommRenPrem * (1 + asmp.CnsmpTax)
    else:
        return 0


def SizeExpsMaint(t):
    """Maintenance expense per policy at time t"""
    return (SizeAnnPrem(t) * asmp.ExpsMaintAnnPrem
            + (SizeSumAssured(t) * asmp.ExpsMaintSA + asmp.ExpsMaintPol)
            * InflFactor(t))


def SizeExpsOther(t):
    """Other expenses per policy at time t"""
    return 0


def SizeInvstIncome(t):
    """Investment Income per policy from t to t+1"""
    return (SizeReserveTotalAftMat(t) + SizePremium(t)) * InvstRetRate(t)


def SizePremium(t):
    """Premium income per policy from t to t+1"""
    return SizeSumAssured(t) * pol.GrossPremRate * pol.PremFreq


def SizeReservePremRsrvAftMat(t):
    """Premium reserve per policy: After maturity"""
    return SizeSumAssured(t) * pol.ReserveNLP_Rate('VAL', t)


def SizeReservePremRsrvEnd(t):
    """Premium reserve per policy: End of period"""
    return SizeSumAssured(t) * pol.ReserveNLP_Rate('VAL', t)


def SizeReserveTotalAftMat(t):
    """Total reserve per policy: After maturity"""
    return (SizeReservePremRsrvAftMat(t)
           + SizeReserveUernPremAftMat(t))


def SizeReserveUernPremAftMat(t):
    """Unearned premium: After maturity"""
    return 0 # SizeSumAssured(t) * polset.UnernPremRate(polset, tt, True)


def SizeReserveUernPremEnd(t):
    """Unearned reserve per policy: End of period"""
    return 0 # SizeSumAssured(t) * pol.UnernPremRate(polset, tt)


def SizeSumAssured(t):
    """Sum assured per policy at time ``t``"""
    return  pol.SumAssured


def last_t():
    return min(asmp.LastAge - pol.IssueAge, pol.PolicyTerm)


def InflFactor(t):
    return scen[ScenID].InflFactor(t)


def InvstRetRate(t):
    return scen[ScenID].InvstRetRate(t)


def SurrRateMult(t):
    """Surrender rate multiple (Default: 1)"""
    return asmp.SurrRateMult(t)


# ---------------------------------------------------------------------------
# References

asmp = ("Interface", (".", "Assumptions"), "auto")

scen = ("Interface", ("..", "Economic"), "auto")


import os.path
import sys
import shutil
import math
import modelx as mx
import lifelib
import lifelib.commands.create as cmd
import pytest


@pytest.mark.parametrize('argv', [
    [],
    ['--template', 'basiclife']
])
def test_argparser(argv):
    tempdir = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                           'mylife')

    parser = cmd.get_argparser()
    argv.append(tempdir)
    args = parser.parse_args(argv)

    assert args.template == 'basiclife' and args.proj_dir == tempdir


@pytest.mark.parametrize('argv', [
    [],
    ['--template', 'basiclife']
])
def test_main(argv, tmp_path):
    tempdir = str(tmp_path / 'mylife')

    argv.append(tempdir)
    cmd.main(argv)

    while True:
        try:
            shutil.rmtree(tempdir)
            break
        except (PermissionError, OSError):
            pass

    assert True


@pytest.mark.parametrize(
    "library, model_dir, target, method, args, expected",
    [
        ["basiclife",
         "BasicTerm_SE",
         ("BasicTerm_SE", "Projection", (1,), "pv_net_cf", ()),
         None,
         None,
         108625.00562388412],

        ["basiclife",
         "BasicTerm_ME",
         ("BasicTerm_ME", "Projection", "pv_net_cf", ()),
         "__getitem__",
         (0,),
         108625.00562388412],

        ["basiclife",
         "BasicTerm_S",
         ("BasicTerm_S", "Projection", (1,), "pv_net_cf", ()),
         None,
         None,
         910.92066093366],

        ["basiclife",
         "BasicTerm_M",
         ("BasicTerm_M", "Projection", "pv_net_cf", ()),
         "__getitem__",
         (0,),
         910.92066093366],

        ["savings",
         "CashValue_SE",
         ("CashValue_SE", "Projection", "pv_net_cf", ()),
         None,
         None,
         4950173.438724911],

        ["savings",
         "CashValue_ME",
         ("CashValue_ME", "Projection", "pv_net_cf", ()),
         "__getitem__",
         (0,),
         4950173.438724911],

        ["fastlife",
         "model",
         ("fastlife", "Projection", "PV_NetCashflow", (0,)),
         "__getitem__",
         (1,),
         8954.018303162458],

        ["simplelife",
         "model",
         ('simplelife', 'Projection', (1,), 'PV_NetCashflow', (0,)),
         None,
         None,
         8954.018303162458],

        ["nestedlife",
         "model",
         ('nestedlife', 'OuterProj', (1,), 'PV_NetCashflow', (0,)),
         None,
         None,
         8954.018303162458],

        ["ifrs17sim",
         "model",
         ('ifrs17sim', 'OuterProj', (1,), 'CSM', (10,)),
         None,
         None,
         1946.6462116677003],

        ["solvency2",
         "model",
         ('solvency2', 'SCR_life', (0, 1), 'SCR_life', ()),
         None,
         None,
         5929.437677484881],

        ["smithwilson",
         "model",
         ('smithwilson', 'SmithWilson', 'P', (10,)),
         None,
         None,
         1.0216540491640598]

    ]
)
def test_package(library, model_dir, target, method, args, expected, tmp_path):

    lib_dir = os.path.join(tmp_path, library)
    lifelib.create(library, lib_dir)

    m = mx.read_model(os.path.join(lib_dir, model_dir))

    actual = mx.core.mxsys.get_object_from_tupleid(target)

    if method:
        actual = getattr(actual, method)(*args)

    assert math.isclose(actual, expected)



import sys
import os.path
import pickle
import pathlib

import pytest

from lifelib.projects.solvency2.scripts import solvency2
from lifelib.tests.data.generate_testdata_solvency2 import generate_data

if '' not in sys.path:
    sys.path.insert(0, '')

datadir = pathlib.Path(__file__).parents[1].joinpath('data')
testdata = str(datadir.joinpath('data_solvency2'))


def test_solvency2():
    model = solvency2.build()
    data = generate_data(model)

    with open(testdata, 'rb') as file:
        data_saved = pickle.load(file)

    assert data == data_saved





"""Module to import additional cells for SCR calculation from input file.

This module contains a function :py:func:`build_input_scr`,
which is called from :py:func:`~solvency2.solvency2.build`,
after :py:func:`~solvency2.build_input`
to create additional cells in ``Input`` space.

"""

def build_input_scr(space, input_file):
    """Create ``CorrLife`` and ``LifeFactors`` cells in ``Input`` space"""

    space.new_cells_from_excel(
            book=input_file,
            range_='CorrLife',
            names_row=0,
            param_cols=[0],
            names_col=0,
            param_rows=[1],
            param_order=[0, 1])
    
    space.new_cells_from_excel(
            book=input_file,
            range_='LifeFactors',
            names_row=0,
            param_cols=[0, 1, 2, 3])
    
    return space



"""
Account value distribution
===========================================

10,000 random numbers drawn from the standard normal distribution
are generated for each time step.
The graph shows how well the 10,000 random numbers for ``t=0``
fit the PDF of the standard normal distribution.

"""
import modelx as mx
import pandas as pd
import matplotlib.pyplot as plt
from scipy.stats import norm, lognorm
import numpy as np

model = mx.read_model("CashValue_ME_EX1")
rand_nums = model.Projection.std_norm_rand()
pv_avs = model.Projection.pv_claims_from_av('MATURITY')
num_bins = 100
S0 = 45000000
sigma = 0.03
T = 10

fig, ax = plt.subplots()
n, bins, patches = ax.hist(rand_nums[:, 0], bins=num_bins, density=True)
ax.plot(bins, norm.pdf(bins), '-')
fig.suptitle('Standard normal distribution for t=0')

#%%
# The distibution of the account value at ``t=120`` follows a log normal distribution.
# In the expression below, :math:`S_{T}` and :math:`S_{0}` denote the account value
# at ``t=T=120`` and ``t=0`` respectively.
#
# .. math::
#
#   \ln\frac{S_{T}}{S_{0}}\sim\phi\left[\left(r-\frac{\sigma^{2}}{2}\right)T, \sigma\sqrt{T}\right]
#
# The graph shows how well the distribution of :math:`e^{-rT}S_{T}`, the present
# values of the account value at `t=0`, fits the PDF of a log normal ditribution.
#
# Reference: *Options, Futures, and Other Derivatives* by John C.Hull
#
# .. seealso::
#
#     * :doc:`/libraries/notebooks/savings/savings_example1` notebook in the :mod:`~savings` library


fig, ax = plt.subplots()
n, bins, patches = ax.hist(pv_avs, bins=num_bins, density=True)
ax.plot(bins, lognorm.pdf(bins, sigma * T**0.5, scale=S0), '-')
fig.suptitle('PV of account value at t=120')






"""
:mod:`solvency2` cashflows
===========================

Net liability cashflows by SCR shock scenarios for selected policies
"""
import modelx as mx
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
sns.set()

model = mx.read_model("model")


def draw(polid, t0):

    fig, ax = plt.subplots()
    fig.suptitle('Net Cashflows')

    title = 'Policy ID: ' + str(polid) + ', Shock at ' + str(t0)

    data = {}
    for risk, shock in (('base', None),
                        ('mort', None),
                        ('longev', None),
                        ('lapse', 'up'),
                        ('lapse', 'down'),
                        ('lapse', 'mass'),
                        ('exps', None)):

        proj = model.SCR_life[t0, polid].Projection(risk, shock)
        last_t = proj.last_t()
        label = risk + ('(' + shock + ')' if shock else '')
        data[label] = [proj.NetInsurCF[t] for t in range(t0, last_t)]

    return pd.DataFrame(
        data, index=range(t0, last_t)).plot(
        ax=ax, kind='line', title=title)


for polid, t0 in [(41, 5), (171, 5)]:
    draw(polid, t0)



import sys
import os.path
import pickle
import pathlib

from lifelib.projects.simplelife.scripts import simplelife
from lifelib.tests.data.generate_testdata import round_signif

if '' not in sys.path:
    sys.path.insert(0, '')

model = simplelife.build(load_saved=False)

testdata = \
    str(pathlib.Path(__file__).parents[1].joinpath('data/data_simplelife'))


def test_simpleflie():
    data = []
    proj = model.Projection
    for i in range(10, 301, 10):
        data.append(round_signif(proj(i).PV_NetCashflow(0), 10))

    with open(testdata, 'rb') as file:
        data_saved = pickle.load(file)

    assert data == data_saved





"""Source module to create ``Economic`` space from.

.. rubric:: Project Templates

This module is included in the following project templates.

* :mod:`simplelife`
* :mod:`nestedlife`

References:
    Scenario

"""

from modelx.serialize.jsonvalues import *

_formula = lambda ScenID: None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def DiscRate(t):
    return Scenarios[ScenID, "IntRate", t]


def InflFactor(t):
    if t == 0:
        return 1
    else:
        return InflFactor(t-1) / (1 + AsmpLookup("InflRate"))


def InvstRetRate(t):
    return DiscRate(t)


# ---------------------------------------------------------------------------
# References

ScenID = 1

AsmpLookup = ("Interface", ("..", "Input", "AsmpLookup"), "auto")

Scenarios = ("Pickle", 2325101521416)


"""Input from files

The ``Input`` Space includes References that hold input values read from
the Excel input file, *input.xlsx*.
It also includes a few Cells to loock up values from the Referneces.

The References in this Space refer to `ExcelRange`_ objects.
The `ExcelRange` objects hold values read from ranges in the input
Excel file, *input.xlsx*. `ExcelRange`_ objects are Python's mapping objects,
so they support most methods that :obj:`dict` has, and can be
converted to :obj:`dict`. The table below lists the References
and their associated named ranges in *input.xlsx*.

==================== ====================
 References            Named ranges
==================== ====================
PolicyData             PolicyData
MortalityTables        MortalityTables
AssumptionTables       AsmpByDuration
Scenarios              Scenarios
DiscountRate           LargePolDiscount
==================== ====================


.. rubric:: References in this Space

Attributes:
    PolicyData: `ExcelRange`_ object holding policy data. The data
        is read from *PolicyData* range in *input.xlsx*.


    ProductSpec: `ExcelRange`_ object holding the data of product specs.
        The data is read from *ProductSpecTable* range in *input.xlsx*.

    Assumption: `ExcelRange`_ object holding the data of the assumption table.
        The data is read from *AssumptionTable* range in *input.xlsx*.

    MortalityTables: `ExcelRange`_ object holding the data of mortality Tables.
        The data is read from *MortalityTables* range in *input.xlsx*.

    AssumptionTables: `ExcelRange`_ object holding the data of assumptions by duration.
        The data is read from *AsmpByDuration* range in *input.xlsx*.

    Scenarios: `ExcelRange`_ object holding the data of economic scenarios.
        The data is read from *Scenarios* range in *input.xlsx*.

    DiscountRate: `ExcelRange`_ object holding the data of premium discount.
        The data is read from *LargePolDiscount* range in *input.xlsx*.


.. _ExcelRange:
   https://docs.modelx.io/en/latest/reference/dataclient.html#excelrange

.. _mapping:
   https://docs.python.org/3/glossary.html#term-mapping

"""

from modelx.serialize.jsonvalues import *

_formula = None

_bases = []

_allow_none = True

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def AsmpLookup(asmp, prod=None, polt=None, gen=None):
    return  Assumption.get((asmp, prod, polt, gen), None)


def SpecLookup(spec, prod=None, polt=None, gen=None):
    return  ProductSpec.get((spec, prod, polt, gen), None)


# ---------------------------------------------------------------------------
# References

PolicyData = ("Pickle", 2325087635976)

MortalityTables = ("Pickle", 2325069517000)

AssumptionTables = ("Pickle", 2325094460872)

Scenarios = ("Pickle", 2325094462088)

DiscountRate = ("Pickle", 2325094616584)

PremWaiverCost = ("Pickle", 2325094616392)

Assumption = ("Pickle", 2325095091912)

ProductSpec = ("Pickle", 2325095091784)


"""
:mod:`fastlife` present values of net cashflows
===============================================
"""

import matplotlib.pyplot as plt
import seaborn as sns
sns.set()
import modelx as mx

proj = mx.read_model("model").Projection
pols = proj.Policy.PolicyData()


for prod in ["TERM", "WL", "ENDW"]:

    fig, ax = plt.subplots()
    fig.suptitle('PV Net Cashflow by Issue Age')
    ax.set_title('Product: ' + prod)
    title = 'Product: ' + prod
    ax.set_xlabel('Issue Age')
    ax.set_ylabel('PV Net Cashflow')

    for sex, marker in zip(["M", "F"], ["o", "^"]):
        filter = (pols.Product == prod) & (
                pols.Sex == sex) & (pols.IssueAge < 60)
        ages = proj.Policy.IssueAge().loc[filter]
        pvcfs = proj.PV_NetCashflow(0).loc[filter]
        ax.scatter(ages, pvcfs, marker=marker)


plt.show()



"""The main module to build a simplelife model.

This module contains only one function :py:func:`build`,
which creates a model from source modules and return it.

If this module is run as a script, the :py:func:`build` function is called
and the created model is available as ``model`` global variable.
"""
import os
import modelx as mx

def build(load_saved=False):
    """Build a model and return it.

    Read input data from `input.xlsm`, create `Input` space and its
    subspace and cells and populate them with the data.

    Args:
        load_saved: If ``True``, input data is read from `simplelife.mx` file
            instead of `input.xlsm`, which is saved when
            :py:func:`build_input <simplelife.build_input.build_input>`
            is executed last time. Defaults to ``False``
    """

    # Make sure the current directory is this folder
    os.chdir(os.path.abspath(os.path.dirname(__file__)))

    # ------------------------------------------------------------------------
    # Build Input space

    from build_input import build_input

    if load_saved:
        model = mx.open_model('simplelife.mx')
        input = model.Input
    else:
        model = mx.new_model(name='simplelife')
        input = build_input(model, 'input.xlsx')
        model.save('simplelife.mx')

    # ------------------------------------------------------------------------
    # Build CommFunc space

    lifetable_refs = {'Input': input}

    def lifetable_params(Sex, IntRate, TableID):
        refs={'MortalityTable': Input.MortalityTables(TableID).MortalityTable}
        return {'refs': refs}

    lifetable = model.import_module(
        module='lifetable',
        name='LifeTable',
        formula=lifetable_params,
        refs=lifetable_refs)

    # ------------------------------------------------------------------------
    # Build Policy space

    from policy import policy_attrs

    policy_refs = {'PolicyData': input.PolicyData,
                   'ProductSpec': input.ProductSpec,
                   'LifeTable': lifetable,
                   'PolicyAttrs': policy_attrs}

    def policy_params(PolicyID):
        refs = {attr: PolicyData[PolicyID].cells[attr]() for attr in PolicyAttrs}
        alias = {'PremTerm': refs['PolicyTerm'],
                 'x': refs['IssueAge'],
                 'm': refs['PolicyTerm'],
                 'n': refs['PolicyTerm']}

        refs.update(alias)
        return {'refs': refs}

    policy = model.import_module(
        module='policy',
        name='Policy',
        formula=policy_params,
        refs=policy_refs)

    # ------------------------------------------------------------------------
    # Build Assumption space

    asmp_refs = {'Policy': policy,
                 'ProductSpec': input.ProductSpec,
                 'MortalityTables': input.MortalityTables,
                 'asmp': input.Assumption,
                 'asmp_tbl': input.AssumptionTables}

    def asmp_params(PolicyID):
        refs = {'pol': Policy[PolicyID]}
        alias = {'prod': refs['pol'].Product,
                 'polt': refs['pol'].PolicyType,
                 'gen': refs['pol'].Gen}
        refs.update(alias)
        return {'refs': refs}

    asmp = model.import_module(
        module='assumption',
        name='Assumption',
        formula=asmp_params,
        refs=asmp_refs)

    asmp.allow_none = True

    # ------------------------------------------------------------------------
    # Build Assumption space

    def econ_params(ScenID):
        refs = {'Scenario': Input.Scenarios[ScenID]}
        return {'refs': refs}

    economic = model.import_module(
        module='economic',
        name='Economic',
        formula=econ_params,
        refs={'asmp': asmp,
              'Input': input})

    # ------------------------------------------------------------------------
    # Build Projection space

    projbase = model.import_module(
        module='projection',
        name='BaseProj')

    pvmixin = model.import_module(
        module='present_value',
        name='PV')

    proj_refs = {'Policy': policy,
                 'Assumption': asmp,
                 'Economic': economic}

    def proj_params(PolicyID, ScenID=1):
        refs = {'pol': Policy[PolicyID],
                'asmp': Assumption[PolicyID],
                'scen': Economic[ScenID],
                'DiscRate': Economic[ScenID].DiscRate}
        return {'refs': refs}

    proj = model.new_space(
        name='Projection',
        bases=[projbase, pvmixin],
        formula=proj_params,
        refs=proj_refs)

    return model


if __name__ == '__main__':
    model = build()







import os.path

_here = os.path.abspath(os.path.dirname(__file__))


TEMPLATE_DIRS = [
    os.path.join(_here, 'libraries'),
    os.path.join(_here, 'projects')
]

# dict of library paths
TEMPLATES = {}
for d in TEMPLATE_DIRS:
    for f in os.listdir(d):
        fullpath = os.path.join(d, f)
        if os.path.isdir(fullpath) and f[0] != '_':
            TEMPLATES[f] = fullpath

DEFAULT_TEMPLATE = "basiclife"




import sys, os, pickle
from lifelib.tests.data.generate_testdata import round_signif
from lifelib.projects.nestedlife.scripts import nestedlife

import numpy as np

if '' not in sys.path:
    sys.path.insert(0, '')

# %% Code block for overriding the default model

def SurrRateMult(t):
    if t == 0:
        return 1
    else:
        return SurrRateMult(t-1)


def PolsSurr(t):
    """Number of policies: Surrender"""
    return PolsIF_Beg1(t) * asmp.SurrRate(t) * SurrRateMult(t)


def PolsIF_End_inner(t):
    """Number of policies: End of period"""
    if t == t0:
        return outer.PolsIF_End(t)
    else:
        return PolsIF_Beg1(t-1) - PolsDeath(t-1) - PolsSurr(t-1)


def set_model(model):
    # Policy point ID and aliases
    polid = 171
    outer = model.OuterProj

    model.BaseProj.new_cells(formula=SurrRateMult)
    model.BaseProj.PolsSurr.set_formula(PolsSurr)
    model.OuterProj.InnerProj.PolsIF_End.set_formula(PolsIF_End_inner)

    outer[polid].SurrRateMult[1] = 2
    outer[polid].SurrRateMult[2] = 0.5
    outer[polid].SurrRateMult[3] = 1

    inner = outer[polid].InnerProj

    inner[1].SurrRateMult[1] = 2
    inner[2].SurrRateMult[2] = 0.5
    inner[3].SurrRateMult[3] = 1

    return outer[polid]


def update_model(model):

    polid = 171
    outer = model.OuterProj

    outer[polid].SurrRateMult[1] = 0.5
    outer[polid].SurrRateMult[2] = 2
    outer[polid].SurrRateMult[3] = 1

    inner = outer[polid].InnerProj

    inner[1].SurrRateMult[1] = 0.5
    inner[2].SurrRateMult[2] = 2
    inner[3].SurrRateMult[3] = 1

    return outer[polid]


def get_nested(outer, item):

    cells = outer.cells[item]

    act = [cells[t] for t in range(50)]
    expect = []

    for t0 in range(0, 6):
        expect_t0 = [np.nan] * 50
        for t in range(0, 50):
            if t < t0:
                expect_t0[t] = 0
            else:
                cells = outer.InnerProj[t0].cells[item]
                expect_t0[t] = cells[t]

        expect.append(expect_t0)

    act = [round_signif(val, 10) for val in act]
    expect = [[round_signif(val, 10) for val in inner] for inner in expect]

    return [act, expect]


def save_data(outer, filename):

    data = get_nested(outer, 'PolsSurr')
    filepath = os.path.join(os.path.dirname(__file__), filename)

    with open(filepath, 'wb') as file:
        pickle.dump(data, file, protocol=4)


def generate_data(model):
    save_data(set_model(model), 'data_nestedlife1')
    save_data(update_model(model), 'data_nestedlife2')


if __name__ == '__main__':
    generate_data(model = nestedlife.build(True))




